/*******************************************************************
 *
 *	File:		Dialogs.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *      Modified for Series 60 by Staffan Ulfberg (staffan@ulfberg.se)
 *
 *	(c) Copyright 2002, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <eikprogi.h>
#include "Dialogs.h"
#include "CSimpleTextParser.h"
#include "emameapp.h"

extern "C"
{
#include "audit.h"
}


#include <txtrich.h>				// CRichText
#include <eikrted.h>				// CEikRichTextEditor
#if defined __SERIES80__ || defined __NOKIA7710__
#include <eikmfne.h>
#endif


/************************************************
 *
 * GameRunning Dialog
 *
 ************************************************/

CGameRunningDialog::CGameRunningDialog(const TDesC8& aGameName,
                                       CMameEngine& aMameEngine)
  : iMameEngine(aMameEngine)
{
  iGameName.Copy(aGameName);
}

void CGameRunningDialog::PreLayoutDynInitL()
{
  CEikLabel *label =(CEikLabel*)Control(ECtlRunningGameName);
  label->SetTextL(iGameName);
}

#if defined __SERIES60__ || defined __SERIES80__ || defined __NOKIA7710__
TBool
CGameRunningDialog::OkToExitL(TInt aCommandId)
{
  switch (aCommandId)
  {
    case ECmdRunningStop:
      iMameEngine.ForceGameExit();
      iMameEngine.ResumeGame();
      break;
      
    case ECmdRunningResume:
      CMameAppUi::Static()->UnblockKeys();
      iMameEngine.ResumeGame();
      break;
    default:
      break;
  }
  return false;
}
#elif defined __UIQ__
void CGameRunningDialog::HandleControlStateChangeL(TInt aControlId)
{
  switch (aControlId)
  {
    case ECtlCmdButPanelRunningGame:
    {
      const TInt ctrlId =
        STATIC_CAST(CEikDialogToolBar*, Control(aControlId))
        ->StateChangeControlId();
      switch (ctrlId)
      {
        case ECtlCmdButStopGame:
          iMameEngine.ForceGameExit();
          iMameEngine.ResumeGame();
          break;
          
        case ECtlCmdButResumeGame:
          iMameEngine.ResumeGame();
          break;
        default:
          break;
      }
      break;
    }
    default:
      break;
  }
}
#endif


/************************************************
 *
 * Scanning ROMs Dialog
 *
 ************************************************/


CScanningRomDialog::CScanningRomDialog(CMameEngine& aMameEngine, 
                                       CInstalledGameList &aGameList)
#ifdef __SERIES60__
  : CAknProgressDialog(&iThis),
#else
  : CEikDialog(),
#endif
    iThis(this),
    iMameEngine(aMameEngine),
    iInstalledGameList(aGameList)
{
}

CScanningRomDialog::~CScanningRomDialog()
{
  delete iIdle;
}

void
CScanningRomDialog::PreLayoutDynInitL()
{
  iIdle = CIdle::NewL(CActive::EPriorityIdle);
  iIdle->Start(TCallBack(&CScanningRomDialog::IdleCallBack, this));
  
  iNoOfGames = 0;
  const struct GameDriver **game = iMameEngine.GameDrivers();
  while(*game++)
  {
    iNoOfGames++;
  }
#ifdef __SERIES60__
  iProgressInfo = GetProgressInfoL();
#else
  iProgressInfo = 
    STATIC_CAST(CEikProgressInfo *, Control(ECtlScanningRomProgress));
#endif
  iProgressInfo->SetFinalValue(iNoOfGames);
}

TInt
CScanningRomDialog::IdleCallBack(TAny* aParam)
{
  return ((CScanningRomDialog*) aParam)->DoUpdate();
}

TInt
CScanningRomDialog::DoUpdate()
{
  TInt missing = !CMameEngine::CheckIfGameRomsComplete(iGameNo);
  if (!missing)
  {
    iInstalledGameList.AddGame(iGameNo);
  }
  
  iGameNo++;
  iProgressInfo->IncrementAndDraw(1);
  
  if (iGameNo == iNoOfGames)
  {
#ifdef __SERIES60__
    TRAPD(err, ProcessFinishedL());
#else
    TRAPD(err, TryExitL(EEikBidOk));
#endif
    return 0;
  }
  return 1;
}

/************************************************
 *
 * Game Info Dialog
 *
 ************************************************/

CGameInfoDialog::CGameInfoDialog(CMameEngine& aMameEngine, TInt aGame) 
  : CSimpleTextInfoDialog(R_GAME_INFO_CLONE),
    iMameEngine(aMameEngine),
    iGame(aGame)
{
}

void CGameInfoDialog::AppendTextL(TDes &aOutput, 
                                  TInt aResId, 
                                  const TDesC &aGameText)
{
  TBuf<256> text;
  HBufC * format = CEikonEnv::Static()->AllocReadResourceLC(aResId);
  text.Format(*format, &aGameText);
  aOutput.Append(text);
  CleanupStack::PopAndDestroy(format);
}

void CGameInfoDialog::ShowTextL(CRichText &aRichText)
{
  TBuf8<64> text8;
  TBuf<64> text;
  TBuf<512> fullText;
		
  const struct GameDriver *game = *(iMameEngine.GameDrivers() + iGame);
  
  text8.Format(_L8("%s (%s)"), game->description, (const TUint8*)game->name);
  text.Copy(text8);
  AppendTextL(fullText, R_GAME_INFO_TITLE, text);
  
  if (iMameEngine.IsGameCloneOf(game))
  {
    text8.Format(_L8("%s"), game->clone_of->description);
    text.Copy(text8);
    AppendTextL(fullText, R_GAME_INFO_CLONE, text);
  }
  
  text8.Format(_L8("%s, %s"), game->year, game->manufacturer);
  text.Copy(text8);
  AppendTextL(fullText, R_GAME_INFO_MANUFACTURED, text);
  
  text8.Format(_L8("%d x %d, using %d colors"), 
               game->drv->screen_width,
               game->drv->screen_height,
               game->drv->total_colors);
  text.Copy(text8);
  AppendTextL(fullText, R_GAME_INFO_RESOLUTION, text);

  for (TInt i = 0; i < MAX_CPU; i++)
  {
    if (game->drv->cpu[i].cpu_type != 0)
    {
      float freq = game->drv->cpu[i].cpu_clock;
      freq /= (1000 * 1000);
      text8.Format(_L8("%s @ %.2f MHz %s"),
                   iMameEngine.CpuName(game->drv->cpu[i].cpu_type), 
                   freq,
                   game->drv->cpu[i].cpu_type & CPU_AUDIO_CPU ? "(sound)" : "");
      text.Copy(text8);
      AppendTextL(fullText, R_GAME_INFO_CPU, text);
    }
  }
  ShowSimpleTextL(fullText, aRichText);
}

/************************************************
 *
 * Game Options Dialog
 *
 ************************************************/

#ifdef __SERIES60__
COptionsListView::COptionsListView(TGameOptions gameOptions)
  : iGameOptions(gameOptions)
{
}

COptionsListView::~COptionsListView()
{
  CEikonEnv* env = CEikonEnv::Static();
  env->EikAppUi()->RemoveFromStack(this);
}

void
COptionsListView::CreateL()
{
  CEikonEnv* env = CEikonEnv::Static();
  ConstructFromResourceL(R_GAME_OPTIONS_DIALOG);
  env->EikAppUi()->AddToStackL(this);
  MakeVisible(ETrue);
  SetRect(env->EikAppUi()->ClientRect());
  ActivateL();
  DrawNow();
}

CAknSettingItem *
COptionsListView::CreateSettingItemL(TInt aIdentifier)
{
  CAknSettingItem *settingItem = NULL;
  switch (aIdentifier)
  {
    case EGameOptionsGameSizeId:
      settingItem = new (ELeave)CAknEnumeratedTextPopupSettingItem(
        aIdentifier, *((TInt*)&iGameOptions.iGameSize));
      break;
    case EGameOptionsKeepAspectRatioToggleId:
      settingItem = new (ELeave)
        CAknBinaryPopupSettingItem(aIdentifier, iGameOptions.iKeepAspectRatio);
      break;
    case EGameOptionsScreenOrientationId:
      settingItem = new (ELeave)CAknEnumeratedTextPopupSettingItem(
        aIdentifier, *((TInt*)&iGameOptions.iScreenOrientation));
      break;
    case EGameOptionsSoundToggleId:
      settingItem = new (ELeave)
        CAknBinaryPopupSettingItem(aIdentifier, iGameOptions.iSoundEnabled);
      break;
    case EGameOptionsSoundVolumeId:
      settingItem = new (ELeave) 
        CAknVolumeSettingItem(aIdentifier, iGameOptions.iSoundVolume);
      break;
    case EGameOptionsFrameSkipId:
      settingItem = new (ELeave) 
        CAknIntegerEdwinSettingItem(aIdentifier, iGameOptions.iFrameSkip);
      break;
    case EGameOptionsAutoFrameSkipToggleId:
      settingItem = new (ELeave)
        CAknBinaryPopupSettingItem(aIdentifier, 
                                   iGameOptions.iAutoFrameSkipEnabled);
      break;
    case EGameOptionsMaxAutoFrameSkipId:
      settingItem = new (ELeave) 
        CAknIntegerEdwinSettingItem(aIdentifier, 
                                    iGameOptions.iMaxAutoFrameSkip);
      break;
    case EGameOptionsThrottleSpeedToggleId:
      settingItem = new (ELeave)
        CAknBinaryPopupSettingItem(aIdentifier, 
                                   iGameOptions.iThrottleSpeedEnabled);
      break;
    default:
      break;
  }
  return settingItem;
}

TGameOptions
COptionsListView::GameOptions() 
{
  return iGameOptions;
}
    
TKeyResponse
COptionsListView::OfferKeyEventL(const TKeyEvent& aKeyEvent, TEventCode aType)
{
  if (aType != EEventKey)
  {
    return EKeyWasNotConsumed;
  }
  TInt index = ListBox()->CurrentItemIndex();
  index = SettingItemArray()->ItemIndexFromVisibleIndex(index);
  CAknSettingItem *item = (*SettingItemArray())[index];
  
  if (item->Identifier()==EGameOptionsSoundVolumeId)
  {
    switch (aKeyEvent.iCode)
    {
      case EKeyRightArrow:
        iGameOptions.iSoundVolume++;
        if (iGameOptions.iSoundVolume > 10)
        {
          iGameOptions.iSoundVolume = 10;
        }
        LoadSettingsL();
        DrawNow();
        break;
      case EKeyLeftArrow:
        iGameOptions.iSoundVolume--;
        if (iGameOptions.iSoundVolume < 1)
        {
          iGameOptions.iSoundVolume = 1;
        }
        LoadSettingsL();
        DrawNow();
        break;
      default:
        CAknSettingItemList::OfferKeyEventL(aKeyEvent, aType);
        break;
    }
  }
  else if (item->Identifier()==EGameOptionsSoundToggleId ||
           item->Identifier()==EGameOptionsGameSizeId ||
           item->Identifier()==EGameOptionsKeepAspectRatioToggleId ||
           item->Identifier()==EGameOptionsScreenOrientationId ||
           item->Identifier()==EGameOptionsAutoFrameSkipToggleId ||
           item->Identifier()==EGameOptionsThrottleSpeedToggleId)
  {
    CAknEnumeratedTextPopupSettingItem *enumItem =
      STATIC_CAST(CAknEnumeratedTextPopupSettingItem*, item);
    switch (aKeyEvent.iCode) 
    {
      case EKeyDevice3: // fire button
      case EKeyEnter:
        enumItem->EditItemL(EFalse);			
        item->StoreL();
        DrawNow();
        break;
      default:
        CAknSettingItemList::OfferKeyEventL(aKeyEvent, aType);
        break;
    }
  }
  else if (item->Identifier()==EGameOptionsFrameSkipId ||
           item->Identifier()==EGameOptionsMaxAutoFrameSkipId)
  {
    CAknIntegerSettingItem *integerItem = 
      STATIC_CAST(CAknIntegerSettingItem*, item);
    switch (aKeyEvent.iCode) 
    {
      case EKeyDevice3: // fire button
      case EKeyEnter:
        integerItem->EditItemL(EFalse);			
        item->StoreL();
        DrawNow();
        break;
      default:
        CAknSettingItemList::OfferKeyEventL(aKeyEvent, aType);
        break;
    }
  }

  return EKeyWasConsumed;
}

void
COptionsListView::SizeChanged()
{
  if (ListBox()) 
  {
    ListBox()->SetRect(Rect());
  }
}

#else

void CGameOptionsDialog::PreLayoutDynInitL()
{
  CEikChoiceList *choiceList = 
    (CEikChoiceList *)Control(EGameOptionsGameSizeId);
  choiceList->SetCurrentItem((TInt)iGameOptions.iGameSize);

  CEikCheckBox *checkBox;
  checkBox = 
    static_cast<CEikCheckBox *>(Control(EGameOptionsKeepAspectRatioToggleId));
  checkBox->SetState(iGameOptions.iKeepAspectRatio ? 
                     CEikButtonBase::ESet : CEikButtonBase::EClear);

  choiceList = (CEikChoiceList *)Control(EGameOptionsScreenOrientationId);
  choiceList->SetCurrentItem((TInt)iGameOptions.iScreenOrientation);

#if defined __UIQ__ || defined __NOKIA7710__
  checkBox =
    static_cast<CEikCheckBox *>(Control(EGameOptionsVirtualKeyboardToggleId));
  checkBox->SetState(iGameOptions.iVirtualKeyboardEnabled ? 
                     CEikButtonBase::ESet : CEikButtonBase::EClear);
#endif

  checkBox = static_cast<CEikCheckBox *>(Control(EGameOptionsSoundToggleId));
  checkBox->SetState(iGameOptions.iSoundEnabled ? 
                     CEikButtonBase::ESet : CEikButtonBase::EClear);

#ifdef __UIQ__
  CQikNumberEditor *numberEditor;
  numberEditor = 
    static_cast<CQikNumberEditor*>(Control(EGameOptionsSoundVolumeId));
  numberEditor->SetValueL(iGameOptions.iSoundVolume);
#else
  CEikNumberEditor *numberEditor;
  numberEditor = 
    static_cast<CEikNumberEditor *>(Control(EGameOptionsSoundVolumeId));
  numberEditor->SetNumber(iGameOptions.iSoundVolume);
#endif

#ifdef __UIQ__
  numberEditor = 
    static_cast<CQikNumberEditor *>(Control(EGameOptionsFrameSkipId));
  numberEditor->SetValueL(iGameOptions.iFrameSkip);
#else
  numberEditor = 
    static_cast<CEikNumberEditor *>(Control(EGameOptionsFrameSkipId));
  numberEditor->SetNumber(iGameOptions.iFrameSkip);
#endif

  checkBox = 
    static_cast<CEikCheckBox *>(Control(EGameOptionsAutoFrameSkipToggleId));
  checkBox->SetState(iGameOptions.iAutoFrameSkipEnabled  ? 
                     CEikButtonBase::ESet : CEikButtonBase::EClear);

#ifdef __UIQ__
  numberEditor = 
    static_cast<CQikNumberEditor *>(Control(EGameOptionsMaxAutoFrameSkipId));
  numberEditor->SetValueL(iGameOptions.iMaxAutoFrameSkip);
#else
  numberEditor = 
    static_cast<CEikNumberEditor *>(Control(EGameOptionsMaxAutoFrameSkipId));
  numberEditor->SetNumber(iGameOptions.iMaxAutoFrameSkip);
#endif

  checkBox = 
    static_cast<CEikCheckBox *>(Control(EGameOptionsThrottleSpeedToggleId));
  checkBox->SetState(iGameOptions.iThrottleSpeedEnabled  ? 
                     CEikButtonBase::ESet : CEikButtonBase::EClear);
}

void CGameOptionsDialog::SetSizeAndPositionL(const TSize& aSize)
{
  TSize size = aSize;
  if (size.iWidth < 400)
    size.iWidth = 400;
  SetCornerAndSize(EHCenterVCenter, size);
}

TBool CGameOptionsDialog::OkToExitL(TInt aKeycode)
{
  CEikChoiceList *choiceList = 
    static_cast<CEikChoiceList *>(Control(EGameOptionsGameSizeId));
  iGameOptions.iGameSize = (TGameOptions::TGameSize)choiceList->CurrentItem();

  CEikCheckBox *checkBox;
  checkBox = 
    static_cast<CEikCheckBox *>(Control(EGameOptionsKeepAspectRatioToggleId));
  iGameOptions.iKeepAspectRatio = 
    (checkBox->State() == CEikButtonBase::ESet) ? ETrue : EFalse;

  choiceList = 
    static_cast<CEikChoiceList *>(Control(EGameOptionsScreenOrientationId));
  iGameOptions.iScreenOrientation = 
    (TGameOptions::TScreenOrientation)choiceList->CurrentItem();

#if defined __UIQ__ || defined __NOKIA7710__
  checkBox = 
    static_cast<CEikCheckBox *>(Control(EGameOptionsVirtualKeyboardToggleId));
  iGameOptions.iVirtualKeyboardEnabled = 
    (checkBox->State() == CEikButtonBase::ESet) ? ETrue : EFalse;
#endif

  checkBox = static_cast<CEikCheckBox *>(Control(EGameOptionsSoundToggleId));
  iGameOptions.iSoundEnabled = 
    (checkBox->State() == CEikButtonBase::ESet) ? ETrue : EFalse;

#ifdef __UIQ__
  CQikNumberEditor *numberEditor;
  numberEditor = 
    static_cast<CQikNumberEditor *>(Control(EGameOptionsSoundVolumeId));
  iGameOptions.iSoundVolume = numberEditor->Value();
#else
  CEikNumberEditor *numberEditor;
  numberEditor = 
    static_cast<CEikNumberEditor *>(Control(EGameOptionsSoundVolumeId));
  iGameOptions.iSoundVolume = numberEditor->Number();
#endif

#ifdef __UIQ__
  numberEditor = 
    static_cast<CQikNumberEditor *>(Control(EGameOptionsFrameSkipId));
  iGameOptions.iFrameSkip = numberEditor->Value();
#else
  numberEditor = 
    static_cast<CEikNumberEditor *>(Control(EGameOptionsFrameSkipId));
  iGameOptions.iFrameSkip = numberEditor->Number();
#endif

  checkBox = 
    static_cast<CEikCheckBox *>(Control(EGameOptionsAutoFrameSkipToggleId));
  iGameOptions.iAutoFrameSkipEnabled =
    (checkBox->State() == CEikButtonBase::ESet) ? ETrue : EFalse;
  
#ifdef __UIQ__
  numberEditor = 
    static_cast<CQikNumberEditor *>(Control(EGameOptionsMaxAutoFrameSkipId));
  iGameOptions.iMaxAutoFrameSkip = numberEditor->Value();
#else
  numberEditor = 
    static_cast<CEikNumberEditor *>(Control(EGameOptionsMaxAutoFrameSkipId));
  iGameOptions.iMaxAutoFrameSkip = numberEditor->Number();
#endif

  checkBox = 
    static_cast<CEikCheckBox *>(Control(EGameOptionsThrottleSpeedToggleId));
  iGameOptions.iThrottleSpeedEnabled =
    (checkBox->State() == CEikButtonBase::ESet) ? ETrue : EFalse;
  return ETrue;
}
#endif

/************************************************
 *
 * SimpleTextInfoDialog
 *
 ************************************************/

CSimpleTextInfoDialog::CSimpleTextInfoDialog(TInt aTextId,
                                             TBool aSimpleTextResIdIsArray)
  : iSimpleTextResId(aTextId),
    iSimpleTextResIdIsArray(aSimpleTextResIdIsArray)
{
}

void CSimpleTextInfoDialog::PreLayoutDynInitL()
{
  CEikRichTextEditor *richTextEditor = 
    STATIC_CAST(CEikRichTextEditor *, Control(ERichTextEditorControlId));
#ifdef __UIQ__ 
  richTextEditor->SetSize(richTextEditor->Size());
  richTextEditor->SetEdwinSizeObserver(this);
#endif
  richTextEditor->CreateScrollBarFrameL();
  richTextEditor->ScrollBarFrame()->
    SetScrollBarVisibilityL(CEikScrollBarFrame::EOff,
                            CEikScrollBarFrame::EAuto);
#if defined __SERIES80__ || defined __NOKIA7710__
  richTextEditor->SetSize(TSize(300, 150));
#endif
}

void
CSimpleTextInfoDialog::PostLayoutDynInitL()
{
  CEikRichTextEditor *richTextEditor = 
    STATIC_CAST(CEikRichTextEditor *, Control(ERichTextEditorControlId));
  ShowTextL(*(richTextEditor->RichText()));
  richTextEditor->UpdateAllFieldsL();
#ifdef __UIQ__
  Layout();
#endif
#ifdef __SERIES60__
  richTextEditor->SetSize(Rect().Size());
#endif
}

TBool
CSimpleTextInfoDialog::HandleEdwinSizeEventL(CEikEdwin* aEdwin, 
                                             TEdwinSizeEvent aEventType,
                                             TSize aDesirableEdwinSize)
{
  if ((aEventType == EEventSizeChanging))
  {
    aEdwin->SetSize(aDesirableEdwinSize);
  }
  return EFalse;
}

void
CSimpleTextInfoDialog::ShowTextL(CRichText &aRichText)
{
  if (iSimpleTextResIdIsArray)
  {
    CSimpleTextFormatParser *parser = CSimpleTextFormatParser::NewLC();
    CDesCArrayFlat *desArray =
      CEikonEnv::Static()->ReadDesCArrayResourceL(iSimpleTextResId);
    CleanupStack::PushL(desArray);
    for (TInt i=0 ; i < desArray->Count(); i++)
    {
      parser->ParseL(desArray->operator[](i), aRichText);
    }
    CleanupStack::PopAndDestroy(desArray);
    CleanupStack::PopAndDestroy(parser);
  }
  else
  {
    HBufC *text = CEikonEnv::Static()->AllocReadResourceLC(iSimpleTextResId);
    ShowSimpleTextL(*text, aRichText);
    CleanupStack::PopAndDestroy(text);
  }
}

void
CSimpleTextInfoDialog::ShowSimpleTextL(const TDesC& aSimpleText,
                                       CRichText& aRichText)
{
  CSimpleTextFormatParser *parser = CSimpleTextFormatParser::NewLC();
  parser->ParseL(aSimpleText, aRichText);
  CleanupStack::PopAndDestroy(parser);
}
