/*******************************************************************
 *
 *	File:		Audio.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *      Modified for Series 60 by Staffan Ulfberg.
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include "audio.h"

#include <f32file.h>
#include <e32svr.h>
#include <e32base.h>
#include <string.h>

/*******************************************
 *
 * CGameAudio
 *
 *******************************************/

CGameAudio::CGameAudio(TBool aStereo) 
  : iStereo(aStereo)
{
}

CGameAudio *CGameAudio::NewL(TBool aStereo)
{
  CGameAudio *self = new(ELeave) CGameAudio(aStereo);
  CleanupStack::PushL(self);
  self->ConstructL();
  CleanupStack::Pop();
  return self;
}

CGameAudio::~CGameAudio()
{
  delete iMdaAudioOutputStream;
  iMdaAudioOutputStream = NULL;
  
  for (TInt i=0 ; i < KMaxSoundBuffers ; i++)
  {
    delete[] iSoundBuffers[i];
    iSoundBuffers[i] = NULL;
  }
}

void CGameAudio::ConstructL()
{
  iMdaAudioOutputStream = CMdaAudioOutputStream::NewL(*this);
  iMdaAudioOutputStream->Open(&iMdaAudioDataSettings);

  WaitForOpenToCompleteL();

  iMdaAudioOutputStream->
    SetAudioPropertiesL(TMdaAudioDataSettings::ESampleRate8000Hz,
                        TMdaAudioDataSettings::EChannelsMono);
  for (TInt i=0; i < KMaxSoundBuffers; i++)
  {
    iSoundBuffers[i] = new TUint16[KBufferedSamples];
  }

  iCurrentBuffer = 0;
  iCurrentPosition = 0;
  iNextBufferToSend = 0;
  iNoOfOutstandingBuffers = 0;
}

TInt 
CGameAudio::Write(TInt16* aBuffer, TInt aSize)
{
  TInt remaining = aSize;
  while (remaining > 0)
  {
    TInt toCopy = remaining;
    if (toCopy > KBufferedSamples - iCurrentPosition)
    {
      toCopy = KBufferedSamples - iCurrentPosition;
    }
    TUint16 *dst = &iSoundBuffers[iCurrentBuffer][iCurrentPosition];
    for (TInt i = toCopy; i > 0; i--)
    {
      *dst++ = *aBuffer++;
      if (iStereo)
      {
        aBuffer++; // we're mono
      }
    }
    iCurrentPosition += toCopy;
    remaining -= toCopy;
    
    if (iCurrentPosition == KBufferedSamples)
    {
      TInt oldestNotCopied =
        (iNextBufferToSend - iNoOfOutstandingBuffers + KMaxSoundBuffers) %
        KMaxSoundBuffers;
      TInt afterCurrent = (iCurrentBuffer + 1) % KMaxSoundBuffers;
      
      if (afterCurrent == oldestNotCopied)
      {
        // We're full.  Throw away remaining data...
        break;
      }
      
      iCurrentBuffer = afterCurrent;
      iCurrentPosition = 0;
      SendBufferIfPossible();
    }
  }
    
  return aSize;
}

void
CGameAudio::SetVolume(TInt volume)
{
  iMdaAudioOutputStream->SetVolume(volume);
}

TInt
CGameAudio::Volume()
{
  return iMdaAudioOutputStream->Volume();
}

TInt
CGameAudio::MaxVolume()
{
  return iMdaAudioOutputStream->MaxVolume();
}

void 
CGameAudio::SendBufferIfPossible()
{
  if (iNextBufferToSend == iCurrentBuffer)
  {
    return;
  }

  if (iNoOfOutstandingBuffers >= KMaxSoundBuffers - 1)
  {
    return;
  }

  iSentBuffers[iNextBufferToSend].
    Set((const TText8 *) iSoundBuffers[iNextBufferToSend], 
        KBufferedSamples * 2);
  iMdaAudioOutputStream->WriteL(iSentBuffers[iNextBufferToSend]);
  iNoOfOutstandingBuffers++;
  iNextBufferToSend = (iNextBufferToSend + 1) % KMaxSoundBuffers;
}

void CGameAudio::WaitForOpenToCompleteL()
{
  TInt count = 20; // 5 seconds
  TInt waitPeriod = 250 * 1000;
  while (!iIsOpen && --count)
  {
    User::After(waitPeriod);
    TInt error;
    CActiveScheduler::RunIfReady(error, CActive::EPriorityIdle);
  }
  if (!iIsOpen)
  {
    User::LeaveIfError(KErrNotSupported);
  }
}

void CGameAudio::MaoscOpenComplete(TInt aError)
{
  if (aError)
  {
    RDebug::Print(_L("CGameAudio::MaoscOpenComplete, error=%d"), aError);
  }
  iIsOpen = ETrue;
}

void CGameAudio::MaoscBufferCopied(TInt aError, const TDesC8& /*aBuffer*/)
{
  iNoOfOutstandingBuffers--;
  ASSERT(iNoOfOutstandingBuffers >= 0);
  SendBufferIfPossible();

  if (aError)
    RDebug::Print(_L("CGameAudio::MaoscBufferCopied, error=%d"), aError);
}

void CGameAudio::MaoscPlayComplete(TInt aError)
{
  if (aError)
    RDebug::Print(_L("CGameAudio::MaoscPlayComplete, error=%d"), aError);
  ASSERT(iNoOfOutstandingBuffers == 0);
}
