/*******************************************************************
 *
 *	File:		OsdEpoc.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *      Modified for Series 60 by Staffan Ulfberg (staffan@ulfberg.se)
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

extern "C" {
#include "osdepend.h"
#include "driver.h"
int osd_mark_vector_dirty(int x, int y);
extern int bitmap_dirty;        /* in mame.c */
}

#undef NULL

#include <e32std.h>
#include <f32file.h>

#include <stdio.h>
#include <stdarg.h>

#include "target.h"
#include "emame.h"
#include "video.h"
#include "audio.h"
#include "EMameGlobal.h"

extern int dos_skip_next_frame(struct osd_bitmap *bitmap);
extern void dos_change_framskip_from_keys(struct osd_bitmap *bitmap);

extern int video_fps;
extern int autoframeskip;
extern int frameskip;
extern int max_autoframeskip;
extern int throttle;

static const TInt sSafetyArea = 16;

static TInt sJoystickX = 0;
static TInt sJoystickY = 0;
static TBool sJoystickPressed = EFalse;


/**********************************
 *
 * Global variables
 *
 **********************************/

extern "C" {
int gForcedExitFromFrontEnd = 0;
}

static TBool sResumeFromFrontEnd = EFalse;
TBool gDirectScreenAccessAborted = EFalse;

TBool getResumeFromFrontEnd()
{
	return sResumeFromFrontEnd;
}

void SetResumeFromFrontEnd(TBool aResumeFromFrontEnd)
{
	sResumeFromFrontEnd = aResumeFromFrontEnd;
}

/******************************************
 *
 * CMameWindow
 *
 ******************************************/

class CMameWindow : public CGameWindow
{
  public:
    ~CMameWindow();
    static CMameWindow* NewL(TBool aVirtualKeyboardEnabled);
    void SetGamePalette(CGamePalette* aGamePalette);
    CGamePalette *GetGamePalette();
    
    void SetGameSize(TGameOptions::TGameSize);
    
  protected:
    void ConstructL(TBool aVirtualKeyboardEnabled);
    CGamePalette*						iMamePalette;
};

CMameWindow* CMameWindow::NewL(TBool aVirtualKeyboardEnabled)
{
  CMameWindow* self = new(ELeave)CMameWindow;
  CleanupStack::PushL(self);
  self->ConstructL(aVirtualKeyboardEnabled);
  CleanupStack::Pop();
  return self;
}

CMameWindow::~CMameWindow()
{
	delete iMamePalette;
}

void CMameWindow::ConstructL(TBool aVirtualKeyboardEnabled)
{
  CGameWindow::ConstructL(_L("EMAME"), aVirtualKeyboardEnabled);
}

void CMameWindow::SetGameSize(TGameOptions::TGameSize aGameSize)
{
  switch (aGameSize)
  {
    case TGameOptions::EGameSizeNormal:
      SetGameScreenDrawer(new CDirectBitmapDrawer(*this));
      break;
    case TGameOptions::EGameSizeFitScreen:
      SetGameScreenDrawer(new CFastScaledBitmapDrawer(*this));
      break;
    case TGameOptions::EGameSizeFitScreenNicely:
      SetGameScreenDrawer(CGammaCorrectScaledBitmapDrawer::NewL(*this));
      break;
  }
}

void
CMameWindow::SetGamePalette(CGamePalette* aGamePalette)
{
  iMamePalette = aGamePalette;
  GameScreenDevice()->SetPalette(iMamePalette);
}

CGamePalette *
CMameWindow::GetGamePalette()
{
  return iMamePalette;
}

/**********************************
 *
 * Static variables
 *
 **********************************/

static CMameWindow *sMameWindow = NULL;
static CGameAudio *sMameAudio = NULL;
static int sMasterVolume = -32;

int osd_init(void)
{
  sMameWindow = NULL;
  sMameAudio = NULL;
  sMasterVolume = -32;
  
  return 0;
}

void osd_exit(void)
{
	DEBUGGER();
}

struct osd_bitmap *
osd_alloc_bitmap(int width, int height, int depth)
{
  ASSERT(depth == 8);	//  for now we only support 8 bpp

  struct osd_bitmap *bitmap = NULL;

  TRAPD(err, 
        bitmap = (struct osd_bitmap *)User::AllocL(sizeof(struct osd_bitmap)));
  if (err)
  {
    logerror("osd_alloc_bitmap failed: alloc osd_bitmap struct");
    return NULL;
  }
  bitmap->width	= width;
  bitmap->height = height;
  bitmap->depth = depth;
  bitmap->_private = 0;
  bitmap->line = 0;

  TInt bytesPerLine = ((depth + 7) / 8) * (width + 2 * sSafetyArea);
  TInt size = bytesPerLine * (height + 2 * sSafetyArea);

  TRAPD(err2, bitmap->_private = User::AllocL(size));
  if (err2)
  {
    logerror("osd_alloc_bitmap failed: alloc osd_bitmap frame buffer");
    osd_free_bitmap(bitmap);
    return NULL;
  }
  Mem::FillZ(bitmap->_private, size);

  TRAPD(err3,
        bitmap->line = (TUint8 **) User::AllocL(height * sizeof(TUint8 *)));
  if (err3)
  {
    logerror("osd_alloc_bitmap failed: alloc osd_bitmap line indices");
    osd_free_bitmap(bitmap);
    return NULL;
  }

  for (TInt i=0; i < height; i++)
  {
    bitmap->line[i] = 
      (TUint8 *)bitmap->_private + bytesPerLine * (i + sSafetyArea) +
      sSafetyArea * ((depth + 7) / 8);
  }

  return bitmap;
}

void
osd_clearbitmap(struct osd_bitmap *bitmap)
{
  int i;
  for (i = 0; i < bitmap->height; i++)
  {
    if (bitmap->depth == 16)
      memset(bitmap->line[i],0,2*bitmap->width);
    else
      memset(bitmap->line[i], 0, bitmap->width);
  }

  if (bitmap == Machine->scrbitmap)
  {
    osd_mark_dirty(0, 0, -1 , -1 , 1);
    bitmap_dirty = 1;
  }
}

void 
osd_free_bitmap(struct osd_bitmap *bitmap)
{
  if (bitmap)
  {
    if (bitmap->_private)
    {
      User::Free(bitmap->_private);
    }
    if (bitmap->line)
    {
      User::Free(bitmap->line);
    }
    User::Free(bitmap);
  }
}

int
osd_create_display(int width,
                   int height,
                   int depth,
                   int fps,
                   int attributes,
                   int orientation)
{
  video_fps = fps;

  TBool virtualKeyboardEnabled = EFalse;
#if defined __UIQ__ || defined __NOKIA7710__
  virtualKeyboardEnabled = GameOptions().iVirtualKeyboardEnabled;
#endif
  TRAPD(err, sMameWindow = CMameWindow::NewL(virtualKeyboardEnabled));
  if (err)
  {
    DPRINT((HERE("osd_create_display: failed, err=%d"), err));
    return err;
  }

  TGameOptions::TGameSize gameSize = GameOptions().iGameSize;
  if ((attributes & VIDEO_TYPE_VECTOR) &&
      gameSize != TGameOptions::EGameSizeNormal)
  {
    gameSize = TGameOptions::EGameSizeNormal;
  }
  sMameWindow->SetGameSize(gameSize);

  sMameWindow->SetKeepAspectRatio(GameOptions().iKeepAspectRatio);
  
  /* 
   * set visible area to nothing just to initialize it - it will
   * be set by the core
   */
  osd_set_visible_area(0, 0, 0, 0);
  
  frameskip = GameOptions().iFrameSkip;
  autoframeskip = GameOptions().iAutoFrameSkipEnabled ? 1 : 0;
  max_autoframeskip = GameOptions().iMaxAutoFrameSkip;
  throttle = GameOptions().iThrottleSpeedEnabled ? 1 : 0;
  
  return 0;
}

void osd_set_visible_area(int min_x, int max_x, int min_y, int max_y)
{
  sMameWindow->OsdSetVisibleArea(min_x, max_x, min_y, max_y);
  /*
   * set_ui_visarea() is called by OsdSetVisibleArea() above.
   */
}

void osd_close_display(void)
{
		/*
		 * Get rid of these bitmaps before closing the display
		 */
	if (Machine->scrbitmap)
	{
		bitmap_free(Machine->scrbitmap);
		Machine->scrbitmap = NULL;
	}
	if (Machine->debug_bitmap)
	{
		osd_free_bitmap(Machine->debug_bitmap);
		Machine->debug_bitmap = NULL;
	}

	if (sMameWindow)
	{
		delete sMameWindow;
		sMameWindow = NULL;
	}
}

CGamePalette* AllocateMameColorsL(TUint totalcolors, const UINT8 *palette, UINT16 *pens, TDisplayMode aDisplayMode)
{
	CGamePalette*	gamePalette = CGamePalette::NewL(aDisplayMode);
	CleanupStack::PushL(gamePalette);

	for (TUint i = 0; i < totalcolors; i++)
	{
		TRgb	color(palette[3 * i], palette[3 * i + 1], palette[3 * i + 2]);
		pens[i] = (UINT16) gamePalette->AllocateEntry(color, totalcolors > (TUint)gamePalette->GetPaletteMaxIndex());
	}

	Machine->uifont->colortable[0] = gamePalette->AllocateEntry(KRgbBlack);
	Machine->uifont->colortable[1] = gamePalette->AllocateEntry(KRgbWhite);
	Machine->uifont->colortable[2] = gamePalette->AllocateEntry(KRgbWhite);
	Machine->uifont->colortable[3] = gamePalette->AllocateEntry(KRgbBlack);

	CleanupStack::Pop();
	return gamePalette;
}


/* palette is an array of 'totalcolors' R,G,B triplets. The function returns */
/* in *pens the pen values corresponding to the requested colors. */
/* If 'totalcolors' is 32768, 'palette' is ignored and the *pens array is filled */
/* with pen values corresponding to a 5-5-5 15-bit palette */

void OsdAllocateColorsL(unsigned int totalcolors,
		const UINT8 *palette,UINT16 *pens,int modifiable,
		const UINT8 *debug_palette,UINT16 *debug_pens)
{
	TDisplayMode displayMode = sMameWindow->GameScreenDevice()->DisplayMode();

	ASSERT(TGameBitmapUtil::GetColorDepth(displayMode) > 8);
        CGamePalette* gamePalette = AllocateMameColorsL(totalcolors, palette, pens, displayMode);
        sMameWindow->SetGamePalette(gamePalette);
}


int osd_allocate_colors(unsigned int totalcolors,
		const UINT8 *palette,UINT16 *pens,int modifiable,
		const UINT8 *debug_palette,UINT16 *debug_pens)
{
	DPRINT((HERE("osd_allocate_colors: total=%d"), totalcolors));

	TRAPD(err, OsdAllocateColorsL(totalcolors, palette, pens, modifiable, debug_palette, debug_pens));
	if (err)
	{
		DPRINT((HERE("OsdAllocateColorsL failed: err=%d"), err));
	}

	return err;
}


void osd_modify_pen(int pen,unsigned char red, unsigned char green, unsigned char blue)
{
  // DEBUGGER();
  CGamePalette *gamePalette = sMameWindow->GetGamePalette();
  if (gamePalette != NULL)
  {
    gamePalette->ModifyPen(pen, red, green, blue);
  }
}

void osd_get_pen(int pen,unsigned char *red, unsigned char *green, unsigned char *blue)
{
//	DEBUGGER();
  CGamePalette *gamePalette = sMameWindow->GetGamePalette();
  if (gamePalette != NULL)
  {
    gamePalette->GetPen(pen, red, green, blue);
  }
}

void osd_mark_dirty(int xmin, int ymin, int xmax, int ymax, int ui)
{
}


int osd_mark_vector_dirty(int x, int y)
{
//	DEBUGGER();
	return 1;
}

static int skip_next_frame = 0;

int osd_skip_this_frame(void)
{
  return skip_next_frame;
}

void
osd_update_video_and_audio(struct osd_bitmap *game_bitmap,
                           struct osd_bitmap *debug_bitmap,
                           int leds_status)
{
  int skip_this_frame = skip_next_frame;
  skip_next_frame = dos_skip_next_frame(game_bitmap);
  
  if (skip_this_frame == 0 && game_bitmap)
  {
    sMameWindow->GameFrameTick(game_bitmap);

    dos_change_framskip_from_keys(game_bitmap);
  }

  // We need to take care of events from the window and audio servers.
  TInt error;
  while (CActiveScheduler::RunIfReady(error, CActive::EPriorityIdle) ||
         gDirectScreenAccessAborted)
  {
    if (gDirectScreenAccessAborted)
    {
      User::After(250 * 1000);
      sMameWindow->TryToStartDrawing(getResumeFromFrontEnd());
    }
  }
  SetResumeFromFrontEnd(EFalse);

#if defined __UIQ__ || defined __NOKIA7710__
  sMameWindow->GetJoyStatus(sJoystickX, sJoystickY, sJoystickPressed);
  switch (GameOptions().iScreenOrientation)
  {
    case TGameOptions::EScreenOrientationRotateRight:
    {
      TInt tmp = sJoystickY;
      sJoystickY = sJoystickX;
      sJoystickX = -tmp;
      break;
    }
    case TGameOptions::EScreenOrientationUpsideDown:
    {
      sJoystickY = -sJoystickY;
      sJoystickX = -sJoystickX;
      break;
    }
    case TGameOptions::EScreenOrientationRotateLeft:
    {
      TInt tmp = sJoystickY;
      sJoystickY = -sJoystickX;
      sJoystickX = tmp;
      break;
    }
  }
#endif
}

void osd_debugger_focus(int debugger_has_focus)
{
	DEBUGGER();
}

void osd_set_gamma(float _gamma)
{
	DEBUGGER();
}

float osd_get_gamma(void)
{
	DEBUGGER();
	return 0.0;
}

void osd_set_brightness(int brightness)
{
	DEBUGGER();
}

int osd_get_brightness(void)
{
	DEBUGGER();
	return -1;
}

void osd_save_snapshot(struct osd_bitmap *bitmap)
{
  save_screen_snapshot(bitmap);
//	DEBUGGER();
}

/*************************
 *
 * Sound
 *
 *************************/

TInt gSamplesPerFrame = 0;

int
osd_start_audio_stream(int stereo)
{
  gSamplesPerFrame = // zero if no sound
    (TInt)(Machine->sample_rate / Machine->drv->frames_per_second);
  if (gSamplesPerFrame != 0)
  {
    TRAPD(err, sMameAudio = CGameAudio::NewL(stereo != 0));
    if (err)
    {
      gSamplesPerFrame = 0;
    }
  }
  if (gSamplesPerFrame != 0)
  {
    sMameAudio->SetVolume(GameOptions().iSoundVolume * 
                          sMameAudio->MaxVolume() / 10);
    sMasterVolume = sMameAudio->Volume() * 32 / sMameAudio->MaxVolume() - 32;
  }
  return gSamplesPerFrame;
}

int osd_update_audio_stream(INT16 *buffer)
{
  TInt samples = 0;
  if (sMameAudio)
  {
    samples = sMameAudio->Write(buffer, gSamplesPerFrame);
  }
  return samples;
}

void osd_stop_audio_stream(void)
{
	DEBUGGER();
	if (sMameAudio)
	{
		delete sMameAudio;
		sMameAudio = NULL;
	}
}

void
osd_set_mastervolume(int attenuation)
{
  sMasterVolume = attenuation;
  if (sMameAudio != NULL) 
  {
    TInt volume = (attenuation + 32) * sMameAudio->MaxVolume() / 32;
    sMameAudio->SetVolume(volume);
  }
}

int
osd_get_mastervolume(void)
{
  return sMasterVolume;
}

void osd_sound_enable(int enable)
{
	DPRINT((HERE("osd_sound_enable, enable=%d"), enable));
}

/* direct access to the Sound Blaster OPL chip */
void osd_opl_control(int chip,int reg)
{
	DEBUGGER();
}

void osd_opl_write(int chip,int data)
{
	DEBUGGER();
}


/***********************
 *
 * Keyboard
 *
 ***********************/

static struct KeyboardInfo keylist[] =
{
#if defined __SERIES60__
  { "Up",		EStdKeyUpArrow,	        KEYCODE_UP },
  { "Down",		EStdKeyDownArrow,	KEYCODE_DOWN },
  { "Left",		EStdKeyLeftArrow,	KEYCODE_LEFT },
  { "Right",		EStdKeyRightArrow,	KEYCODE_RIGHT },
  { "Select",		EStdKeyDevice3,		KEYCODE_ENTER },
  { "LeftSoft",		EStdKeyDevice0,		KEYCODE_TAB },
  { "RightSoft",	EStdKeyDevice1,	        KEYCODE_ESC },
  { "Del",		EStdKeyBackspace,	KEYCODE_DEL },
  { "0",                '0',                    KEYCODE_0 },
  { "1",                '1',                    KEYCODE_1 },
  { "2",                '2',                    KEYCODE_2 },
  { "3",                '3',                    KEYCODE_3 },
  { "4",                '4',                    KEYCODE_4 },
  { "5",		'5',	                KEYCODE_5 },
  { "6",                '6',                    KEYCODE_6 },
  { "7",		'7',	                KEYCODE_7 },
  { "8",                '8',                    KEYCODE_8 },
  { "9",                '9',                    KEYCODE_9 },
  { "*",                '*',                    KEYCODE_ASTERISK },
  { "#",                EStdKeyHash,            KEYCODE_TILDE },
#elif defined __UIQ__
  { "Up",               EStdKeyDevice4,         KEYCODE_UP },
  { "Down",             EStdKeyDevice5,         KEYCODE_DOWN },
  { "Left",             EStdKeyDevice6,         KEYCODE_LEFT },
  { "Right",            EStdKeyDevice7,         KEYCODE_RIGHT },
  { "Select",           EStdKeyDevice8,         KEYCODE_ENTER }, 
  { "Shortcut",         EStdKeyApplication1,    KEYCODE_TAB },
  { "Speaker",          EStdKeyDeviceA,         KEYCODE_ESC },
  { "GameA",            EStdKeyApplicationA,    KEYCODE_A },
  { "GameB",            EStdKeyApplicationB,    KEYCODE_B },
  { "Browse",           EStdKeyApplication0,    KEYCODE_C },
  { "Green",            EStdKeyYes,             KEYCODE_D },
  { "Red",              EStdKeyNo,              KEYCODE_E },
  { "Vol+",             EStdKeyIncVolume,       KEYCODE_PLUS_PAD },
  { "Vol-",             EStdKeyDecVolume,       KEYCODE_MINUS_PAD },
  { "Record",           EStdKeyDictaphoneRecord,KEYCODE_F },
#elif defined __SERIES80__
  { "Up",		EStdKeyUpArrow,	        KEYCODE_UP },
  { "Down",		EStdKeyDownArrow,	KEYCODE_DOWN },
  { "Left",		EStdKeyLeftArrow,	KEYCODE_LEFT },
  { "Right",		EStdKeyRightArrow,	KEYCODE_RIGHT },
  { "Enter",		EStdKeyEnter,		KEYCODE_ENTER },
  { "Joy Left",		EStdKeyDevice6,	        KEYCODE_4_PAD },
  { "Joy Right",	EStdKeyDevice7,	        KEYCODE_6_PAD },
  { "Joy Up",		EStdKeyDevice8,	        KEYCODE_8_PAD },
  { "Joy Down",		EStdKeyDevice9, 	KEYCODE_2_PAD },
  { "Joy Press",	EStdKeyDeviceA,		KEYCODE_5_PAD },
  { "Tab",		EStdKeyTab,		KEYCODE_TAB },
  { "Esc",	        EStdKeyEscape,	        KEYCODE_ESC },
  { "A",                'A',                    KEYCODE_A },
  { "B",                'B',                    KEYCODE_B },
  { "C",                'C',                    KEYCODE_C },
  { "D",                'D',                    KEYCODE_D },
  { "E",                'E',                    KEYCODE_E },
  { "F",                'F',                    KEYCODE_F },
  { "G",                'G',                    KEYCODE_G },
  { "H",                'H',                    KEYCODE_H },
  { "I",                'I',                    KEYCODE_I },
  { "J",                'J',                    KEYCODE_J },
  { "K",                'K',                    KEYCODE_K },
  { "L",                'L',                    KEYCODE_L },
  { "M",                'M',                    KEYCODE_M },
  { "N",                'N',                    KEYCODE_N },
  { "O",                'O',                    KEYCODE_O },
  { "P",                'P',                    KEYCODE_P },
  { "Q",                'Q',                    KEYCODE_Q },
  { "R",                'R',                    KEYCODE_R },
  { "S",                'S',                    KEYCODE_S },
  { "T",                'T',                    KEYCODE_T },
  { "U",                'U',                    KEYCODE_U },
  { "V",                'V',                    KEYCODE_V },
  { "W",                'W',                    KEYCODE_W },
  { "X",                'X',                    KEYCODE_X },
  { "Y",                'Y',                    KEYCODE_Y },
  { "Z",                'Z',                    KEYCODE_Z },
  { "0",                '0',                    KEYCODE_0 },
  { "1",                '1',                    KEYCODE_1 },
  { "2",                '2',                    KEYCODE_2 },
  { "3",                '3',                    KEYCODE_3 },
  { "4",                '4',                    KEYCODE_4 },
  { "5",                '5',                    KEYCODE_5 },
  { "6",                '6',                    KEYCODE_6 },
  { "7",                '7',                    KEYCODE_7 },
  { "8",                '8',                    KEYCODE_8 },
  { "9",                '9',                    KEYCODE_9 },
  { "=",                EStdKeyEquals,          KEYCODE_EQUALS },
  { "#",                EStdKeyHash,            KEYCODE_TILDE },
  { "Caps",             EStdKeyCapsLock,        KEYCODE_CAPSLOCK },
  { ";",                EStdKeySemiColon,       KEYCODE_COLON },
  { "'",                EStdKeySingleQuote,     KEYCODE_QUOTE },
  { "LShift",           EStdKeyLeftShift,       KEYCODE_LSHIFT },
  { "/",                EStdKeyForwardSlash,    KEYCODE_SLASH },
  { "RShift",           EStdKeyRightShift,      KEYCODE_RSHIFT },
  { "LCtrl",            EStdKeyLeftCtrl,        KEYCODE_LCONTROL },
  { "Chr",              EStdKeyLeftFunc,        KEYCODE_LALT },
  { ",",                EStdKeyComma,           KEYCODE_COMMA },
  { ".",                EStdKeyFullStop,        KEYCODE_STOP },
  { "Space",            EStdKeySpace,           KEYCODE_SPACE },
  { "Menu",             EStdKeyMenu,            KEYCODE_MENU },
#elif defined __NOKIA7710__
  { "Up",		EStdKeyUpArrow,	        KEYCODE_UP },
  { "Down",		EStdKeyDownArrow,	KEYCODE_DOWN },
  { "Left",		EStdKeyLeftArrow,	KEYCODE_LEFT },
  { "Right",		EStdKeyRightArrow,	KEYCODE_RIGHT },
  { "Select",           EStdKeyDevice7,         KEYCODE_ENTER },
  { "GameA",            EStdKeyDevice4,         KEYCODE_A },
  { "GameB",            EStdKeyDevice5,         KEYCODE_B },
  { "Esc",	        EStdKeyEscape,	        KEYCODE_ESC },
#else
#error No target defined
#endif
#if defined __UIQ__ || defined __NOKIA7710__
  { "Virtual Up",       EMameVirtKeyUp,         KEYCODE_8_PAD },
  { "Virtual Down",     EMameVirtKeyDown,       KEYCODE_2_PAD },
  { "Virtual Left",     EMameVirtKeyLeft,       KEYCODE_4_PAD },
  { "Virtual Right",    EMameVirtKeyRight,      KEYCODE_6_PAD },
  { "Virtual Center",   EMameVirtKeyCenter,     KEYCODE_5_PAD },
  { "Virtual Fire",     EMameVirtKeyFire,       KEYCODE_0_PAD },
  { "Virtual Conf",     EMameVirtKeyConf,       KEYCODE_ASTERISK },
  { "Virtual Sel",      EMameVirtKeySel,        KEYCODE_ENTER_PAD },
  { "Virtual Esc",      EMameVirtKeyEsc,        KEYCODE_DEL_PAD },
  { "Virtual Coin",     EMameVirtKeyCoin,       KEYCODE_HOME },
  { "Virtual Start",    EMameVirtKeyStart,      KEYCODE_END },
#endif
  { 0, 0, 0 }	/* end of table */
};

struct inputport_default_keys
{
    UINT32 type;
    InputSeq seq;
};

static struct inputport_default_keys default_keys[] =
{
#if defined __SERIES60__
  { IPT_JOYSTICK_UP | IPF_PLAYER1,               SEQ_DEF_1(KEYCODE_UP) },
  { IPT_JOYSTICK_DOWN | IPF_PLAYER1,             SEQ_DEF_1(KEYCODE_DOWN) },
  { IPT_JOYSTICK_LEFT | IPF_PLAYER1,             SEQ_DEF_1(KEYCODE_LEFT) },
  { IPT_JOYSTICK_RIGHT | IPF_PLAYER1,            SEQ_DEF_1(KEYCODE_RIGHT) },
  { IPT_PADDLE | IPF_PLAYER1,                    SEQ_DEF_1(KEYCODE_LEFT) },
  {(IPT_PADDLE | IPF_PLAYER1)+IPT_EXTENSION,     SEQ_DEF_1(KEYCODE_RIGHT) },
  { IPT_DIAL | IPF_PLAYER1,                      SEQ_DEF_1(KEYCODE_LEFT) },
  {(IPT_DIAL | IPF_PLAYER1)+IPT_EXTENSION,       SEQ_DEF_1(KEYCODE_RIGHT) },
  { IPT_TRACKBALL_X | IPF_PLAYER1,               SEQ_DEF_1(KEYCODE_LEFT) },
  {(IPT_TRACKBALL_X | IPF_PLAYER1)+IPT_EXTENSION,SEQ_DEF_1(KEYCODE_RIGHT) },
  { IPT_TRACKBALL_Y | IPF_PLAYER1,               SEQ_DEF_1(KEYCODE_UP) },
  {(IPT_TRACKBALL_Y | IPF_PLAYER1)+IPT_EXTENSION,SEQ_DEF_1(KEYCODE_DOWN) },
  { IPT_AD_STICK_X | IPF_PLAYER1,                SEQ_DEF_1(KEYCODE_LEFT) },
  {(IPT_AD_STICK_X | IPF_PLAYER1)+IPT_EXTENSION, SEQ_DEF_1(KEYCODE_RIGHT) },
  { IPT_AD_STICK_Y | IPF_PLAYER1,                SEQ_DEF_1(KEYCODE_UP) },
  {(IPT_AD_STICK_Y | IPF_PLAYER1)+IPT_EXTENSION, SEQ_DEF_1(KEYCODE_DOWN) },
  { IPT_UI_UP,                                   SEQ_DEF_1(KEYCODE_UP) },
  { IPT_UI_DOWN,                                 SEQ_DEF_1(KEYCODE_DOWN) },
  { IPT_UI_LEFT,                                 SEQ_DEF_1(KEYCODE_LEFT) },
  { IPT_UI_RIGHT,                                SEQ_DEF_1(KEYCODE_RIGHT) },
#elif defined __SERIES80__ || defined __UIQ__ || defined __NOKIA7710__
  { IPT_JOYSTICK_UP | IPF_PLAYER1,               
    SEQ_DEF_3(KEYCODE_UP, CODE_OR, KEYCODE_8_PAD) },
  { IPT_JOYSTICK_DOWN | IPF_PLAYER1,             
    SEQ_DEF_3(KEYCODE_DOWN, CODE_OR, KEYCODE_2_PAD) },
  { IPT_JOYSTICK_LEFT | IPF_PLAYER1,             
    SEQ_DEF_3(KEYCODE_LEFT, CODE_OR, KEYCODE_4_PAD) },
  { IPT_JOYSTICK_RIGHT | IPF_PLAYER1,            
    SEQ_DEF_3(KEYCODE_RIGHT, CODE_OR, KEYCODE_6_PAD) },
  { IPT_PADDLE | IPF_PLAYER1,                    
    SEQ_DEF_3(KEYCODE_LEFT, CODE_OR, KEYCODE_4_PAD) },
  {(IPT_PADDLE | IPF_PLAYER1)+IPT_EXTENSION,     
   SEQ_DEF_3(KEYCODE_RIGHT, CODE_OR, KEYCODE_6_PAD) },
  { IPT_DIAL | IPF_PLAYER1,                      
    SEQ_DEF_3(KEYCODE_LEFT, CODE_OR, KEYCODE_4_PAD) },
  {(IPT_DIAL | IPF_PLAYER1)+IPT_EXTENSION,       
   SEQ_DEF_3(KEYCODE_RIGHT, CODE_OR, KEYCODE_6_PAD) },
  { IPT_TRACKBALL_X | IPF_PLAYER1,               
    SEQ_DEF_3(KEYCODE_LEFT, CODE_OR, KEYCODE_4_PAD) },
  {(IPT_TRACKBALL_X | IPF_PLAYER1)+IPT_EXTENSION,
   SEQ_DEF_3(KEYCODE_RIGHT, CODE_OR, KEYCODE_6_PAD) },
  { IPT_TRACKBALL_Y | IPF_PLAYER1,               
    SEQ_DEF_3(KEYCODE_UP, CODE_OR, KEYCODE_8_PAD) },
  {(IPT_TRACKBALL_Y | IPF_PLAYER1)+IPT_EXTENSION,
   SEQ_DEF_3(KEYCODE_DOWN, CODE_OR, KEYCODE_2_PAD) },
  { IPT_AD_STICK_X | IPF_PLAYER1,                
    SEQ_DEF_3(KEYCODE_LEFT, CODE_OR, KEYCODE_4_PAD) },
  {(IPT_AD_STICK_X | IPF_PLAYER1)+IPT_EXTENSION, 
   SEQ_DEF_3(KEYCODE_RIGHT, CODE_OR, KEYCODE_6_PAD) },
  { IPT_AD_STICK_Y | IPF_PLAYER1,                
    SEQ_DEF_3(KEYCODE_UP, CODE_OR, KEYCODE_8_PAD) },
  {(IPT_AD_STICK_Y | IPF_PLAYER1)+IPT_EXTENSION, 
   SEQ_DEF_3(KEYCODE_DOWN, CODE_OR, KEYCODE_2_PAD) },
  { IPT_UI_UP,                                   
    SEQ_DEF_3(KEYCODE_UP, CODE_OR, KEYCODE_8_PAD) },
  { IPT_UI_DOWN,                                 
    SEQ_DEF_3(KEYCODE_DOWN, CODE_OR, KEYCODE_2_PAD) },
  { IPT_UI_LEFT,                                 
    SEQ_DEF_3(KEYCODE_LEFT, CODE_OR, KEYCODE_4_PAD) },
  { IPT_UI_RIGHT,                                
    SEQ_DEF_3(KEYCODE_RIGHT, CODE_OR, KEYCODE_6_PAD) },
#endif
#if defined __SERIES60__
  { IPT_UI_CONFIGURE,                            SEQ_DEF_1(KEYCODE_TAB) },
  { IPT_UI_CANCEL,                               SEQ_DEF_1(KEYCODE_ESC) },
  { IPT_UI_SELECT,         SEQ_DEF_3(KEYCODE_ENTER, CODE_OR, KEYCODE_5) },
  { IPT_BUTTON1 | IPF_PLAYER1,                   SEQ_DEF_1(KEYCODE_5) },
  { IPT_BUTTON2 | IPF_PLAYER1,                   SEQ_DEF_1(KEYCODE_7) },
  { IPT_COIN1,	                                 SEQ_DEF_1(KEYCODE_2) },
  { IPT_START1,                                  SEQ_DEF_1(KEYCODE_1) },
  { IPT_UI_PAUSE,		                 SEQ_DEF_1(KEYCODE_0) },
  { IPT_UI_SHOW_FPS,                             SEQ_DEF_1(KEYCODE_TILDE) },
  { IPT_UI_ON_SCREEN_DISPLAY,                    SEQ_DEF_1(KEYCODE_ASTERISK) },
  { IPT_UI_PAN_UP,         SEQ_DEF_2(KEYCODE_UP, KEYCODE_5) },
  { IPT_UI_PAN_DOWN,       SEQ_DEF_2(KEYCODE_DOWN, KEYCODE_5) },
  { IPT_UI_PAN_LEFT,       SEQ_DEF_2(KEYCODE_LEFT, KEYCODE_5) },
  { IPT_UI_PAN_RIGHT,      SEQ_DEF_2(KEYCODE_RIGHT, KEYCODE_5) },
#elif defined __UIQ__
  { IPT_UI_CONFIGURE,
    SEQ_DEF_3(KEYCODE_TAB, CODE_OR, KEYCODE_ASTERISK) },
  { IPT_UI_CANCEL,                               
    SEQ_DEF_3(KEYCODE_ESC, CODE_OR, KEYCODE_DEL_PAD) },
  { IPT_UI_SELECT,                               
    SEQ_DEF_3(KEYCODE_ENTER, CODE_OR, KEYCODE_ENTER_PAD) },
  { IPT_BUTTON1 | IPF_PLAYER1,                   
    SEQ_DEF_3(KEYCODE_A, CODE_OR, KEYCODE_0_PAD) },
  { IPT_BUTTON2 | IPF_PLAYER1,                   SEQ_DEF_1(KEYCODE_B) },
  { IPT_COIN1,	                                 
    SEQ_DEF_3(KEYCODE_C, CODE_OR, KEYCODE_HOME) },
  { IPT_START1,                                  
    SEQ_DEF_3(KEYCODE_D, CODE_OR, KEYCODE_END) },
  { IPT_UI_SHOW_FPS,                             SEQ_DEF_1(KEYCODE_PLUS_PAD)},
  { IPT_UI_ON_SCREEN_DISPLAY,                    SEQ_DEF_1(KEYCODE_MINUS_PAD)},
  { IPT_UI_PAN_UP,         SEQ_DEF_2(KEYCODE_UP, KEYCODE_A) },
  { IPT_UI_PAN_DOWN,       SEQ_DEF_2(KEYCODE_DOWN, KEYCODE_A) },
  { IPT_UI_PAN_LEFT,       SEQ_DEF_2(KEYCODE_LEFT, KEYCODE_A) },
  { IPT_UI_PAN_RIGHT,      SEQ_DEF_2(KEYCODE_RIGHT, KEYCODE_A) },
#elif defined __SERIES80__
  { IPT_UI_CONFIGURE,                            SEQ_DEF_1(KEYCODE_TAB) },
  { IPT_UI_CANCEL,                               SEQ_DEF_1(KEYCODE_ESC) },
  { IPT_UI_SELECT,                               
    SEQ_DEF_3(KEYCODE_ENTER, CODE_OR, KEYCODE_5_PAD) },
  { IPT_BUTTON1 | IPF_PLAYER1,                   SEQ_DEF_1(KEYCODE_LCONTROL) },
  { IPT_BUTTON2 | IPF_PLAYER1,                   SEQ_DEF_1(KEYCODE_LALT) },
  { IPT_BUTTON3 | IPF_PLAYER1,                   SEQ_DEF_1(KEYCODE_SPACE) },
  { IPT_BUTTON4 | IPF_PLAYER1,                   SEQ_DEF_1(KEYCODE_LSHIFT) },
  { IPT_JOYSTICK_UP | IPF_PLAYER2,               SEQ_DEF_1(KEYCODE_R)    },
  { IPT_JOYSTICK_DOWN | IPF_PLAYER2,             SEQ_DEF_1(KEYCODE_F)  },
  { IPT_JOYSTICK_LEFT | IPF_PLAYER2,             SEQ_DEF_1(KEYCODE_D)  },
  { IPT_JOYSTICK_RIGHT | IPF_PLAYER2,            SEQ_DEF_1(KEYCODE_G) },
  { IPT_BUTTON1 | IPF_PLAYER2,                   SEQ_DEF_1(KEYCODE_A) },
  { IPT_BUTTON2 | IPF_PLAYER2,                   SEQ_DEF_1(KEYCODE_S) },
  { IPT_BUTTON3 | IPF_PLAYER2,                   SEQ_DEF_1(KEYCODE_Q) },
  { IPT_BUTTON4 | IPF_PLAYER2,                   SEQ_DEF_1(KEYCODE_W) },
  { IPT_COIN1,                                   SEQ_DEF_1(KEYCODE_5) },
  { IPT_COIN2,                                   SEQ_DEF_1(KEYCODE_6) },
  { IPT_COIN3,                                   SEQ_DEF_1(KEYCODE_7) },
  { IPT_COIN4,                                   SEQ_DEF_1(KEYCODE_8) },
  { IPT_START1,                                  SEQ_DEF_1(KEYCODE_1) },
  { IPT_START2,                                  SEQ_DEF_1(KEYCODE_2) },
  { IPT_START3,                                  SEQ_DEF_1(KEYCODE_3) },
  { IPT_START4,                                  SEQ_DEF_1(KEYCODE_4) },
  { IPT_SERVICE1,                                SEQ_DEF_1(KEYCODE_9) },
  { IPT_SERVICE2,                                SEQ_DEF_1(KEYCODE_0) },
  { IPT_SERVICE3,                                SEQ_DEF_1(KEYCODE_EQUALS) },
  { IPT_SERVICE4,                                SEQ_DEF_1(KEYCODE_DEL) },
  { IPT_UI_PAUSE,                                SEQ_DEF_1(KEYCODE_P) },
  { IPT_UI_SHOW_FPS,                             SEQ_DEF_1(KEYCODE_QUOTE) },
  { IPT_UI_ON_SCREEN_DISPLAY,                    SEQ_DEF_1(KEYCODE_TILDE) },
  { IPT_UI_PAN_UP,                               SEQ_DEF_1(KEYCODE_UP) },
  { IPT_UI_PAN_DOWN,                             SEQ_DEF_1(KEYCODE_DOWN) },
  { IPT_UI_PAN_LEFT,                             SEQ_DEF_1(KEYCODE_LEFT) },
  { IPT_UI_PAN_RIGHT,                            SEQ_DEF_1(KEYCODE_RIGHT) },
  { IPT_UI_FRAMESKIP_DEC,                        SEQ_DEF_1(KEYCODE_I) },
  { IPT_UI_FRAMESKIP_INC,                        SEQ_DEF_1(KEYCODE_O) },
  { IPT_UI_THROTTLE,                             SEQ_DEF_1(KEYCODE_L) },
#elif defined __NOKIA7710__
  { IPT_UI_CONFIGURE,                            SEQ_DEF_1(KEYCODE_ASTERISK) },
  { IPT_UI_CANCEL,                               
    SEQ_DEF_3(KEYCODE_ESC, CODE_OR, KEYCODE_DEL_PAD) },
  { IPT_UI_SELECT,                               
    SEQ_DEF_3(KEYCODE_ENTER, CODE_OR, KEYCODE_ENTER_PAD) },
  { IPT_BUTTON1 | IPF_PLAYER1,                   
    SEQ_DEF_3(KEYCODE_A, CODE_OR, KEYCODE_0_PAD) },
  { IPT_BUTTON2 | IPF_PLAYER1,                   SEQ_DEF_1(KEYCODE_B) },
  { IPT_COIN1,                                   SEQ_DEF_1(KEYCODE_HOME) },
  { IPT_START1,                                  SEQ_DEF_1(KEYCODE_END) },
  { IPT_UI_PAN_UP,         SEQ_DEF_2(KEYCODE_UP, KEYCODE_A) },
  { IPT_UI_PAN_DOWN,       SEQ_DEF_2(KEYCODE_DOWN, KEYCODE_A) },
  { IPT_UI_PAN_LEFT,       SEQ_DEF_2(KEYCODE_LEFT, KEYCODE_A) },
  { IPT_UI_PAN_RIGHT,      SEQ_DEF_2(KEYCODE_RIGHT, KEYCODE_A) },
#endif
  { IPT_END,                                     SEQ_DEF_0 }
};


const struct KeyboardInfo *osd_get_key_list(void)
{
	return keylist;
}

static TInt sRotatedScreenKeyTranslations[][4] =
{
#if defined __SERIES60__ || defined __NOKIA7710__
  {
    EStdKeyUpArrow, EStdKeyDownArrow, EStdKeyLeftArrow, EStdKeyRightArrow
  },  // unrotated
  {
    EStdKeyLeftArrow, EStdKeyRightArrow, EStdKeyDownArrow, EStdKeyUpArrow
  },
  {
    EStdKeyDownArrow, EStdKeyUpArrow, EStdKeyRightArrow, EStdKeyLeftArrow
  },
  {
    EStdKeyRightArrow, EStdKeyLeftArrow, EStdKeyUpArrow, EStdKeyDownArrow
  }
#elif defined __UIQ__
  {
    EStdKeyDevice4, EStdKeyDevice5, EStdKeyDevice6, EStdKeyDevice7
  },  // unrotated
  {
    EStdKeyDevice6, EStdKeyDevice7, EStdKeyDevice5, EStdKeyDevice4
  },
  {
    EStdKeyDevice5, EStdKeyDevice4, EStdKeyDevice7, EStdKeyDevice6
  },
  {
    EStdKeyDevice7, EStdKeyDevice6, EStdKeyDevice4, EStdKeyDevice5
  }
#elif defined __SERIES80__
  {
    EStdKeyDevice8, EStdKeyDevice9, EStdKeyDevice6, EStdKeyDevice7
  },  // unrotated
  {
    EStdKeyDevice6, EStdKeyDevice7, EStdKeyDevice9, EStdKeyDevice8
  },
  {
    EStdKeyDevice9, EStdKeyDevice8, EStdKeyDevice7, EStdKeyDevice6
  },
  {
    EStdKeyDevice7, EStdKeyDevice6, EStdKeyDevice8, EStdKeyDevice9
  }
#endif
};

int
osd_is_key_pressed(int keycode)
{
  TInt orientation = (TInt)GameOptions().iScreenOrientation;
  if (orientation > 0 && orientation <= 3)
  {
    for (int i = 0; i < 4; i++)
    {
      if (sRotatedScreenKeyTranslations[0][i] == keycode)
      {
        keycode = sRotatedScreenKeyTranslations[orientation][i];
        break;
      }
    }
  }

  return sMameWindow ? sMameWindow->IsKeyPressed(keycode) : 0;
}

int osd_readkey_unicode(int flush)
{
	DEBUGGER();
	return 0;
}

#define OS_JOYCODE_UP 0
#define OS_JOYCODE_DOWN 1
#define OS_JOYCODE_LEFT 2
#define OS_JOYCODE_RIGHT 3
#define OS_JOYCODE_BUTTON 4

static struct JoystickInfo joylist[] =
{
#if defined __UIQ__ || defined __NOKIA7710__
  { "Pointer Up",	        OS_JOYCODE_UP,	        JOYCODE_1_UP },
  { "Pointer Down",	        OS_JOYCODE_DOWN,	JOYCODE_1_DOWN },
  { "Pointer Left",	        OS_JOYCODE_LEFT,	JOYCODE_1_LEFT },
  { "Pointer Right",	        OS_JOYCODE_RIGHT,	JOYCODE_1_RIGHT },
  { "Pointer Pressed",          OS_JOYCODE_BUTTON,      JOYCODE_1_BUTTON1 },
#endif
  { 0, 0, 0 }	/* end of table */
};

const struct JoystickInfo *osd_get_joy_list(void)
{
	return joylist;
}

int 
osd_is_joy_pressed(int joycode)
{
  switch (joycode)
  {
    case OS_JOYCODE_UP:
      return sJoystickY < -50 ;
      break;
    case OS_JOYCODE_DOWN:
      return sJoystickY > 50;
      break;
    case OS_JOYCODE_LEFT:
      return sJoystickX < -50;
      break;
    case OS_JOYCODE_RIGHT:
      return sJoystickX > 50;
      break;
    case OS_JOYCODE_BUTTON:
      return sJoystickPressed;
    default:
      ASSERT(EFalse);
  }
  return 0;
}

int
osd_joystick_needs_calibration (void)
{
  return 0;
}

void
osd_joystick_start_calibration (void)
{
}

char *
osd_joystick_calibrate_next (void)
{
  return (char *)NULL;
}

void
osd_joystick_calibrate (void)
{
}

void
osd_joystick_end_calibration (void)
{
}

void
osd_trak_read(int player,int *deltax,int *deltay)
{
  *deltax = *deltay = 0;
}

/* return values in the range -128 .. 128 (yes, 128, not 127) */
void osd_analogjoy_read(int player,int *analog_x, int *analog_y)
{
  *analog_x = sJoystickX;
  *analog_y = sJoystickY;
}

void
osd_customize_inputport_defaults(struct ipd *defaults)
{
  struct inputport_default_keys *i;
  while (defaults->type != IPT_END)
  {
    for (i = default_keys; i->type != IPT_END; i++)
    {
      if (defaults->type == i->type)
      {
        seq_copy(&defaults->seq, &i->seq);
        break;
      }
    }
    if (i->type == IPT_END)
    {
      seq_set_0(&defaults->seq);
    }
    defaults++;
  }
}

void CLIB_DECL logerror(const char *text,...)
{
#if 0
  va_list marker;

  FILE *f = fopen("c:\\stdout.txt", "a");
  fprintf(f, "ERROR: ");
  va_start( marker, text);
  vfprintf(f, text, marker);
  fclose(f);
#endif
}


/* control keyboard leds or other indicators */
void osd_led_w(int led,int on)
{
	DEBUGGER();
}


/*
To start profiling a certain section, e.g. video:
osd_profiler(OSD_PROFILE_VIDEO);
to end profiling of the current section:
osd_profiler(OSD_PROFILE_END);
*/

void osd_profiler(int type)
{
	DEBUGGER();
}

void 
osd_pause(int paused)
{
}

#ifdef MAME_NET
/* network */
int osd_net_init(void){return 0;}
int osd_net_send(int player, unsigned char buf[], int *size){return 0;}
int osd_net_recv(int player, unsigned char buf[], int *size){return 0;}
int osd_net_sync(void){return 0;}
int osd_net_input_sync(void){return 0;}
int osd_net_exit(void){return 0;}
int osd_net_add_player(void){return 0;}
int osd_net_remove_player(int player){return 0;}
int osd_net_game_init(void){return 0;}
int osd_net_game_exit(void){return 0;}
#endif /* MAME_NET */


int
stricmp(const char *string1, const char *string2)
{
  return strcasecmp(string1, string2);
}

unsigned int
osd_cycles()
{
  return User::TickCount();
}
