/*******************************************************************
 *
 *	File:		Video.cpp
 *
 *	Original author: Peter van Sebille (peter@yipton.net)
 *
 *      Heavily modified for Series 60 by Staffan Ulfberg (staffan@ulfberg.se):
 *      - Bitmap scaling routine improved.
 *      - Supports games that modify their palettes.
 *      - Uses CDirectScreenAccess instead of RDirectScreenAccess
 *      - Lots of P800 code removed (easier for me to undertand that way...)
 *      - Lots of reorganization.
 *
 *******************************************************************/

#include <e32math.h>
#include <gdi.h>
#include <stdlib.h>
#include <coedef.h>		// for ECoeWinPriorityNeverAtFront
#include <bitdev.h>
#include "video.h" 

extern "C"
{
#include "memory.h"
#include "input.h"
#include "inptport.h"
#include "osdepend.h"
extern void set_ui_visarea (int xmin, int ymin, int xmax, int ymax);
extern int osd_is_key_pressed(int);
}

extern TBool gDirectScreenAccessAborted;

/*****************************************
 *
 * TGameBitmapUtil implementation
 *
 *****************************************/

TUint8* TGameBitmapUtil::PixelAddress(const CGameScreenDevice* aSrc, const TPoint& aPoint)
{
	return aSrc->ScanLineAddress(aPoint.iY) + ((aPoint.iX * GetBpp(aSrc->DisplayMode())) / 8);
}

/*
 * GetBpp returns the number of bits we need per pixel in the given
 * display mode. It is different from GetColorDepth in that GetBpp can be
 * used to calculate byte offsets in a bitmap/framebuffer.
 * For example, GetBpp(EColor4K) returns 16 because each pixel requires
 * 16 bits (of which only the least significant 12 bits are used).
 * GetColorDepth(EColor4K) returns 12.
 */
TInt TGameBitmapUtil::GetBpp(TDisplayMode aDisplayMode)
{
	const TInt bpp[] =
		{
			-1,
			1,
			2,
			4,
			8,
			4,
			8,
			16,
			32,
			32,
			16,
                        32,
		};
	return bpp[aDisplayMode];
}

TInt TGameBitmapUtil::GetColorDepth(TDisplayMode aDisplayMode)
{
	const TInt bpp[] =
		{
			-1,
			1,
			2,
			4,
			8,
			4,
			8,
			16,
			24,
			32,
			12,
                        24
		};
	return bpp[aDisplayMode];
}



TUint TGameBitmapUtil::GetColorValue(const TRgb& aColor, const TDisplayMode& aDisplayMode)
{
	switch (aDisplayMode)
	{
		case EColor256:
			return aColor.Color256();
			break;
		case EColor4K:
			return aColor.Color4K();
			break;
		case EColor16M:
			return aColor.Color16M();
			break;
#ifdef __HAS_COLOR_16MU__
		case EColor16MU:
			return aColor.Color16MU();
			break;
#endif
		case EColor64K:
			return aColor.Color64K();
			break;
		case ERgb:
			return aColor.Value();
			break;
		default:
			ASSERT(EFalse);
	}
	return 0;
}

TRgb
TGameBitmapUtil::GetRgb(TUint aColor, const TDisplayMode& aDisplayMode)
{
	switch (aDisplayMode)
	{
		case EColor256:
			return TRgb::Color256(aColor);
			break;
		case EColor4K:
			return TRgb::Color4K(aColor);
			break;
		case EColor16M:
			return TRgb::Color16M(aColor);
			break;
#ifdef __HAS_COLOR_16MU__
		case EColor16MU:
			return TRgb::Color16MU(aColor);
			break;
#endif
		case EColor64K:
			return TRgb::Color64K(aColor);
			break;
		default:
			ASSERT(EFalse);
	}
	return 0;
}


/*****************************************
 *
 * CGamePalette implementation
 *
 *****************************************/

CGamePalette::CGamePalette(TDisplayMode aDisplayMode) : iDisplayMode(aDisplayMode)
{
}

CGamePalette::~CGamePalette()
{
}

CGamePalette* CGamePalette::NewL(TDisplayMode aDisplayMode)
{
	CGamePalette*	self = new(ELeave)CGamePalette(aDisplayMode);
	CleanupStack::PushL(self);
	self->ConstructL();
	CleanupStack::Pop();
	return self;
}


void CGamePalette::ConstructL()
{
	iPaletteMaxIndex = 256;
	Mem::FillZ(iPalette, sizeof(iPalette));
	iPaletteLastIndexUsed = 0;
}



TInt CGamePalette::AllocateEntry(const TRgb& aColor, TBool aTryToSaveSpace)
{
	TUint	colorValue = TGameBitmapUtil::GetColorValue(aColor, iDisplayMode);
        if (aTryToSaveSpace)
          for (TInt i=0 ; i<iPaletteLastIndexUsed; i++)
            if (iPalette[i] == colorValue)
              return i;
	ASSERT(iPaletteLastIndexUsed<iPaletteMaxIndex);
	iPalette[iPaletteLastIndexUsed] = colorValue;
	return (iPaletteLastIndexUsed++);
}

void
CGamePalette::ModifyPen(int pen,
                        unsigned char red,
                        unsigned char green,
                        unsigned char blue)
{
  ASSERT(pen < iPaletteMaxIndex);
  TRgb color(red, green, blue);
  TUint	colorValue = TGameBitmapUtil::GetColorValue(color, iDisplayMode);
  iPalette[pen] = colorValue;
}

void
CGamePalette::GetPen(int pen,
                     unsigned char *red,
                     unsigned char *green,
                     unsigned char *blue)
{
  ASSERT(pen < iPaletteMaxIndex);
  TRgb rgb(TGameBitmapUtil::GetRgb(iPalette[pen], iDisplayMode));
  *red = (unsigned char)rgb.Red();
  *green = (unsigned char)rgb.Green();
  *blue = (unsigned char)rgb.Blue();
}

TInt
CGamePalette::GetPaletteMaxIndex()
{
  return iPaletteMaxIndex;
}



/*****************************************
 *
 * CGameScreenDevice implementation
 *
 *****************************************/

CGameScreenDevice::CGameScreenDevice(const TSize& aSize, TDisplayMode aDisplayMode) : iSize(aSize), iDisplayMode(aDisplayMode)
{
}

CGameScreenDevice::~CGameScreenDevice()
{
}

void CGameScreenDevice::Clear()
{
	TInt	height = iSize.iHeight;
	for (TInt h=0 ; h<height ; h++)
	{
		TUint8*		line = ScanLineAddress(h);
		Mem::FillZ(line, BytesPerLine());
	}
}

/*****************************************
 *
 * CFbsGameScreenDevice implementation
 *
 *****************************************/

CFbsGameScreenDevice::CFbsGameScreenDevice(
  const TSize& aSize, TDisplayMode aDisplayMode) : CGameScreenDevice(aSize, aDisplayMode)
{
  iBytesPerLine = CFbsBitmap::ScanLineLength(iSize.iWidth, iDisplayMode);
  // iBytesPerLine = ((TGameBitmapUtil::GetBpp(aDisplayMode) * aSize.iWidth ) + 7) / 8;
}

CFbsGameScreenDevice::~CFbsGameScreenDevice()
{
  delete iBitmap;
  iBitmap = NULL;
}


CFbsGameScreenDevice* CFbsGameScreenDevice::NewL(
  const TSize& aSize, TDisplayMode aDisplayMode)
{
  CFbsGameScreenDevice*	self = new(ELeave) CFbsGameScreenDevice(aSize, aDisplayMode);
  CleanupStack::PushL(self);
  self->ConstructL();
  CleanupStack::Pop();
  return self;
}

TUint8* CFbsGameScreenDevice::ScanLineAddress(TInt aLine) const
{
  iBitmap->LockHeap();
  TUint8 *address = ((TUint8 *)iBitmap->DataAddress()) + aLine * iBytesPerLine;
  iBitmap->UnlockHeap();  // We need to postpone this until we are done painting!
  return address;
}

void CFbsGameScreenDevice::ConstructL()
{
  iBitmap = new (ELeave) CFbsBitmap;
  iBitmap->Create(iSize, iDisplayMode);
}


/*****************************************
 *
 * CHwFbGameScreenDevice implementation
 *
 *****************************************/

CHwFbGameScreenDevice *
CHwFbGameScreenDevice::NewL(const TSize& aWindowSize,
                            TDisplayMode aDisplayMode)
{
  TScreenInfoV01 screenInfo;
  TPckg<TScreenInfoV01> sI(screenInfo);
  UserSvr::ScreenInfo(sI);
  if (!screenInfo.iScreenAddressValid)
  {
    User::Leave(KErrNotSupported);
  }

  /*
   * Note that aWindowSize may differ from screenInfo.iScreenSize.
   * screenInfo.iScreenSize reflects the video controller's configured pixel dimensions, 
   * whereas aWindowSize is the LCD's screen's maximim dimension.
   */
  CHwFbGameScreenDevice *self =
    new(ELeave) CHwFbGameScreenDevice(aWindowSize, aDisplayMode);
  CleanupStack::PushL(self);
  self->ConstructL(screenInfo);
  CleanupStack::Pop();
  return self;
}


CHwFbGameScreenDevice::CHwFbGameScreenDevice(const TSize &aWindowSize,
                                             TDisplayMode aDisplayMode)
  : CGameScreenDevice(aWindowSize, aDisplayMode)
{
}

void CHwFbGameScreenDevice::ConstructL(const TScreenInfoV01& aScreenInfo)
{
  iFrameBuffer = (TUint8*) aScreenInfo.iScreenAddress;
  iBytesPerLine = ((TGameBitmapUtil::GetBpp(iDisplayMode) *
                    aScreenInfo.iScreenSize.iWidth ) + 7) / 8;
  iFrameBuffer += 32; // For 12 or 16 bpp!
}


TUint8* CHwFbGameScreenDevice::ScanLineAddress(TInt aLine) const
{
	return iFrameBuffer + (BytesPerLine() * aLine);
}

/***********************************************
 *
 * CGameWindow
 *
 ***********************************************/

CGameWindow::CGameWindow()
  : iNumberOfCapturedKeys(0),
    iNumberOfCapturedKeyUpAndDowns(0)
{
}

CGameWindow::~CGameWindow()
{
  delete iGameController;
  iGameController = NULL;

  DestructScreenDevice();
  DestructWsResources();
  delete iGameScreenDrawer;
  iGameScreenDrawer = NULL;
}

CGameWindow* CGameWindow::NewL(const TDesC& aGameName,
                               TBool aVirtualKeyboardEnabled)
{
	CGameWindow*	self = new(ELeave)CGameWindow;
	CleanupStack::PushL(self);
	self->ConstructL(aGameName, aVirtualKeyboardEnabled);
	CleanupStack::Pop();
	return self;
}

void CGameWindow::ConstructL(const TDesC& aGameName,
                             TBool aVirtualKeyboardEnabled)
{
	iGameName = aGameName.Left(iGameName.MaxLength());

	ConstructWsResourcesL();
	ConstructScreenDeviceL();

        iGameController = CGameController::NewL(iWsSession, 
                                                *this,
                                                aVirtualKeyboardEnabled);
	ClearScreen();
}

void CGameWindow::Show()
{
  iWsWindowGroup.SetOrdinalPosition(0, 0);
  iWsWindow.SetVisible(ETrue);
  CaptureKeyEvents();
}

void CGameWindow::Hide()
{
  iWsWindowGroup.SetOrdinalPosition(-1, 0);
  /*
   * On UIQ, the window covers the short cut bar at the top of the
   * screen if not made invisible.  On the Series 60, OTOH, we want
   * the window to be visible in case we switch back to it, so that
   * direct screen access can continue.
   */
#ifdef __UIQ__
  iWsWindow.SetVisible(EFalse);
#endif
  ReleaseKeyEvents();
}

void
CGameWindow::CaptureKeyEvents()
{
#ifdef __UIQ__
  /*
   * Trying to capture up and downs from these keys does not work, but
   * we do it anyway, since it might stop some side effects of the
   * keys on some devices...
   */
  CaptureKey(EKeyDeviceA);      // Speaker
  CaptureKeyUpAndDowns(EStdKeyDeviceA);

  CaptureKey(EKeyDictaphoneRecord);      // Record
  CaptureKeyUpAndDowns(EStdKeyDictaphoneRecord);

  CaptureKey(EKeyApplication0); // Shortcut
  CaptureKeyUpAndDowns(EStdKeyApplication0);

  CaptureKey(EKeyApplication1); // Browse
  CaptureKeyUpAndDowns(EStdKeyApplication1);

  CaptureKey(EKeyYes);
  CaptureKeyUpAndDowns(EStdKeyYes);

  CaptureKey(EKeyNo);
  CaptureKeyUpAndDowns(EStdKeyNo);

  CaptureKey(EKeyIncVolume);
  CaptureKeyUpAndDowns(EStdKeyIncVolume);

  CaptureKey(EKeyDecVolume);
  CaptureKeyUpAndDowns(EStdKeyDecVolume);
#endif
}

void
CGameWindow::CaptureKey(TUint aKeyCode)
{
  if (iNumberOfCapturedKeys == KMaxCapturedKeys)
  {
    User::Leave(-1); // What code?
  }
  iCapturedKeys[iNumberOfCapturedKeys] =
    iWsWindowGroup.CaptureKey(aKeyCode, 0, 0);

  User::LeaveIfError(iCapturedKeys[iNumberOfCapturedKeys]);

  iNumberOfCapturedKeys++;
}

void
CGameWindow::CaptureKeyUpAndDowns(TUint aScanCode)
{
  if (iNumberOfCapturedKeyUpAndDowns == KMaxCapturedKeys)
  {
    User::Leave(-1); // What code?
  }
  iCapturedKeyUpAndDowns[iNumberOfCapturedKeyUpAndDowns] =
    iWsWindowGroup.CaptureKeyUpAndDowns(aScanCode, 0, 0);

  User::LeaveIfError(iCapturedKeys[iNumberOfCapturedKeyUpAndDowns]);

  iNumberOfCapturedKeyUpAndDowns++;
}

void
CGameWindow::ReleaseKeyEvents()
{
  for (TInt i = 0; i < iNumberOfCapturedKeys; i++)
  {
    iWsWindowGroup.CancelCaptureKey(iCapturedKeys[i]);
  }
  iNumberOfCapturedKeys = 0;

  for (TInt i = 0; i < iNumberOfCapturedKeyUpAndDowns; i++)
  {
    iWsWindowGroup.CancelCaptureKeyUpAndDowns(iCapturedKeyUpAndDowns[i]);
  }
  iNumberOfCapturedKeyUpAndDowns = 0;
}

void CGameWindow::ConstructWsResourcesL()
{
	TInt	error;

	error = iWsSession.Connect();
	User::LeaveIfError(error);
	iWsScreen=new(ELeave) CWsScreenDevice(iWsSession);
	User::LeaveIfError(iWsScreen->Construct());
	User::LeaveIfError(iWsScreen->CreateContext(iWindowGc));

	iWsWindowGroup=RWindowGroup(iWsSession);
	User::LeaveIfError(iWsWindowGroup.Construct((TUint32)&iWsWindowGroup));
	iWsWindowGroup.SetName(iGameName);

	iWsWindow=RWindow(iWsSession);
	User::LeaveIfError(iWsWindow.Construct(iWsWindowGroup, (TUint32)&iWsWindow));
	iWsWindow.Activate();
	iWsWindow.SetSize(iWsScreen->SizeInPixels());
        
        iWsWindowGroup.DisableKeyClick(ETrue);
#if defined __UIQ__ || defined __NOKIA7710__
        iWsWindow.SetPointerGrab(ETrue);
        iWsWindow.PointerFilter(EPointerFilterDrag, 0);
#endif

        /*
         * Tell the Window Server not to mess about with our process priority
         * Also, because of the way legacy games are written, they never sleep
         * and thus never voluntarily yield the CPU. We set our process priority
         * to EPriorityForeground and hope that a Telephony application on
         * this device runs at EPriorityForeground as well. If not, tough! ;-)
         */
	iWsSession.ComputeMode(RWsSession::EPriorityControlDisabled);
	RProcess		me;
	me.SetPriority(EPriorityForeground);
}


void CGameWindow::DestructWsResources()
{
	if (iWsWindow.WsHandle())
		iWsWindow.Close();

	if (iWsWindowGroup.WsHandle())
		iWsWindowGroup.Close();

	delete iWindowGc;
	iWindowGc = NULL;

	delete iWsScreen;
	iWsScreen = NULL;

	if (iWsSession.WsHandle())
		iWsSession.Close();
}

void CGameWindow::ConstructScreenDeviceL()
{
	TDisplayMode	mode = iWsScreen->DisplayMode();
	TSize			size = iWsScreen->SizeInPixels();

//#ifdef __WINS__
        iUseWServ = ETrue;
//#endif

	if (iUseWServ)
	{
			/*
			 * For WINS we emulate using a Fbs bitmap which we blit to the window's device
			 * context. You can test most behaviour by:
			 * 1) adjusting the size and display mode in the CFbsGameScreenDevice::NewL
			 * 2) configuring the emulator for the appropriate size and mode
			 */
          iGameScreenDevice = CFbsGameScreenDevice::NewL(size, mode);
	}
	else
	{
          iGameScreenDevice = CHwFbGameScreenDevice::NewL(size, mode);
	}

        iGameDirectScreenAccess = CGameDirectScreenAccess::NewL(*this);
}


void CGameWindow::DestructScreenDevice()
{
	delete iGameScreenDevice;
	iGameScreenDevice = NULL;

	delete iGameDirectScreenAccess;
	iGameDirectScreenAccess = NULL;
}

void CGameWindow::GameFrameTick(struct osd_bitmap *bitmap)
{
  iGameScreenDrawer->DrawFrame(bitmap);
#if defined __UIQ__ || defined __NOKIA7710__
  iGameController->DrawVirtualKeyboard();
#endif
  
  if (iUseWServ)
  {
    WsBlitBitmap((CFbsGameScreenDevice*) iGameScreenDevice,  
                 TRect(iGameScreenDevice->Size()));
  }
#ifdef __SERIES60__
  else
  {
    iGameDirectScreenAccess->Update();
  }
#endif
}

void
CGameWindow::TryToStartDrawing(TBool aResumeFromFrontEnd)
{
  iGameDirectScreenAccess->TryToStartDrawing(aResumeFromFrontEnd);
}

void CGameWindow::ClearScreen()
{
	iGameScreenDevice->Clear();
}

TBool
CGameWindow::IsKeyPressed(TInt aKeyCode)
{
  return iGameController->IsKeyPressed(aKeyCode);
}

#if defined __UIQ__ || defined __NOKIA7710__
void
CGameWindow::GetJoyStatus(TInt &x, TInt &y, TBool &pressed)
{
  return iGameController->GetJoyStatus(x, y, pressed);
}
#endif

void CGameWindow::SetGameScreenDrawer(CGameScreenDrawer* aGameScreenDrawer)
{
  iGameScreenDrawer = aGameScreenDrawer;
}

void CGameWindow::WsBlitBitmap(CFbsGameScreenDevice* aFbsGameScreenDevice, const TRect& aRect)
{
	iWindowGc->Activate(iWsWindow);

	iWsWindow.Invalidate(aRect);
	iWsWindow.BeginRedraw(aRect);

	if (aFbsGameScreenDevice)
	{
		iWindowGc->BitBlt(aRect.iTl, aFbsGameScreenDevice->FbsBitmap(), aRect);
	}

	iWsWindow.EndRedraw();
	iWindowGc->Deactivate();
	iWsSession.Flush();
}


/*****************************************
 *
 * CGameController
 *
 *****************************************/

CGameController* CGameController::NewL(RWsSession& aWsSession,
                                       CGameWindow &aGameWindow,
                                       TBool aVirtualKeyboardEnabled)
{
  CGameController *self = new(ELeave)CGameController(aWsSession,
                                                     aGameWindow,
                                                     aVirtualKeyboardEnabled);
  CleanupStack::PushL(self);
  self->ConstructL();
  CleanupStack::Pop();
  return self;
}

CGameController::~CGameController()
{
  Cancel();
#if defined __UIQ__ || defined __NOKIA7710__
  delete iVirtualKeyboardBitmap;
  delete[] iPixelToVirtualButtonMap;
#endif
}

CGameController::CGameController(RWsSession& aWsSession,
                                 CGameWindow &aGameWindow,
                                 TBool aVirtualKeyboardEnabled)
  : CActive(CActive::EPriorityHigh),
    iWsSession(aWsSession),
    iGameWindow(aGameWindow),
    iNoOfKeyPresses(0),
    iX(0),
    iY(0)
#if defined __UIQ__ || defined __NOKIA7710__
  , iVirtualKeyboardEnabled(aVirtualKeyboardEnabled),
    iVirtualKeyboardVisible(EFalse),
    iCurrentPressedRawVirtualKey(ERawVirtKeyNone),
    iPressOccurredInsideVirtualKeyboard(EFalse),
    iPressOccurredInsideCurrentVirtualKey(EFalse)
#endif
{
}

void CGameController::ConstructL()
{
  CActiveScheduler::Add(this);
  iStatus = KRequestPending;
  iWsSession.EventReady(&iStatus);
  SetActive();

#if defined __UIQ__ || defined __NOKIA7710__
  if (iVirtualKeyboardEnabled)
  {
    iVirtualKeyboardBitmap = new (ELeave)CFbsBitmap();
    _LIT(fileName, "c:\\System\\apps\\eemame\\vkeys.mbm");
    iVirtualKeyboardBitmap->Load(fileName, 0);
    
    CFbsBitmap *virtualKeyboardAreaMap = new (ELeave)CFbsBitmap();  
    virtualKeyboardAreaMap->Load(fileName, 1);
    CFbsBitmap *virtualKeyboardAreaColorMap = new (ELeave)CFbsBitmap();  
    virtualKeyboardAreaColorMap->Load(fileName, 2);
    
    iVirtualKeyboardSize = iVirtualKeyboardBitmap->SizeInPixels(); 
    ASSERT(virtualKeyboardAreaMap->SizeInPixels() == iVirtualKeyboardSize);
    ASSERT(virtualKeyboardAreaColorMap->SizeInPixels() == 
           TSize(KNoOfVirtualKeys, 1));
    
    iVirtualKeyboardPosition.iX = 0;
    iVirtualKeyboardPosition.iY =
      iGameWindow.GameScreenDevice()->Size().iHeight - 
      iVirtualKeyboardSize.iHeight;
    
    TInt pixels = iVirtualKeyboardSize.iWidth * iVirtualKeyboardSize.iHeight;
    
    iPixelToVirtualButtonMap = new TUint8[pixels];
    
    TUint16 *areaMap = (TUint16 *)virtualKeyboardAreaMap->DataAddress();
    TUint16 *colorMap = (TUint16 *)virtualKeyboardAreaColorMap->DataAddress();
    ASSERT(ERawVirtKeyNone == KNoOfVirtualKeys);
    for (TInt i = 0; i < pixels; i++)
    {
      TUint16 color = *areaMap++;
      if (color == 0)
      {
        iPixelToVirtualButtonMap[i] = ERawVirtKeyNone;
      }
      else
      {
        TInt j;
        for (j = 0; j < KNoOfVirtualKeys; j++)
        {
          if (color == colorMap[j])
          {
            iPixelToVirtualButtonMap[i] = j;
            break;
          }
        }
        if (j == KNoOfVirtualKeys)
        {
          ASSERT(EFalse);
        }
      }
    }
    delete virtualKeyboardAreaColorMap;
    delete virtualKeyboardAreaMap;

    iVirtualKeyboardVisible = ETrue;
  }
#endif
}

void
CGameController::RunL()
{
  TWsEvent event;
  iWsSession.GetEvent(event);
  if ((event.Type() == EEventKeyDown) || 
      (event.Type() == EEventKey) || 
      (event.Type() == EEventKeyUp))
  {
    HandleKeyEvent(event.Type(), *event.Key());
  }
#if defined __UIQ__ || defined __NOKIA7710__
  else if (event.Type() == EEventPointer)
  {
    HandlePointerEvent(*event.Pointer());
  }
#endif

  iStatus = KRequestPending;
  iWsSession.EventReady(&iStatus);
  SetActive();
}

void
CGameController::DoCancel()
{
  iWsSession.EventReadyCancel();
}

void
CGameController::HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent)
{
  // We use EEventKey since not all keys generate a down event,
  // and EEventKeyDown since modifiers do not generate EEventKey events
  if (aType == EEventKey && aKeyEvent.iRepeats == 0
#ifdef __SERIES80__
      || aType == EEventKeyDown
#endif
    )
  {
    KeyDown(aKeyEvent.iScanCode);
  }
  else if (aType == EEventKeyUp)
  {
    KeyUp(aKeyEvent.iScanCode);
  }
}

#if defined __UIQ__ || defined __NOKIA7710__
void
CGameController::HandlePointerEvent(const TPointerEvent &aPointerEvent)
{
  if (iVirtualKeyboardEnabled)
  {
    if (HandlePossibleVirtualKeyboardPress(aPointerEvent))
    {
      return;
    }
  }

  switch (aPointerEvent.iType)
  {
    case TPointerEvent::EButton1Up:
      iPressed = EFalse;
      break;
    case TPointerEvent::EButton1Down:
      iPressed = ETrue;
    case TPointerEvent::EDrag:
    {
      TRect usedScreenArea = iGameWindow.GetUsedScreenArea();
      iX = (aPointerEvent.iPosition.iX - usedScreenArea.iTl.iX) 
        * 257 / usedScreenArea.Width() - 128;
      iY = (aPointerEvent.iPosition.iY - usedScreenArea.iTl.iY) 
        * 257 / usedScreenArea.Height() - 128;
      if (iY < -128) iY = -128;
      if (iY > 128) iY = 128;
      if (iX < -128) iX = -128;
      if (iX > 128) iX = 128;
      break;
    }
    default:
      break;
  }
}

/*
 * Returns true if the pointer event was consumed.
 */
TBool
CGameController::HandlePossibleVirtualKeyboardPress(
  const TPointerEvent &aPointerEvent)
{
  TBool insideVirtualKeyboard = 
    TRect(iVirtualKeyboardPosition, iVirtualKeyboardSize).
    Contains(aPointerEvent.iPosition);

  if (aPointerEvent.iType == TPointerEvent::EButton1Down)
  {
    iPressOccurredInsideVirtualKeyboard = insideVirtualKeyboard;
    if (!iVirtualKeyboardVisible && insideVirtualKeyboard)
    {
      iVirtualKeyboardVisible = ETrue;
      return ETrue; // Do not let the keyboard handle this event
    }
  }

  if (aPointerEvent.iType == TPointerEvent::EButton1Up ||
      !insideVirtualKeyboard)
  {
    iCurrentPressedRawVirtualKey = ERawVirtKeyNone;
  }
  else
  {
    TRawVirtualKey rawVirtKey = (TRawVirtualKey)iPixelToVirtualButtonMap[
      (aPointerEvent.iPosition.iY - iVirtualKeyboardPosition.iY) *
      iVirtualKeyboardSize.iWidth +
      (aPointerEvent.iPosition.iX - iVirtualKeyboardPosition.iX)];
    
    if (rawVirtKey != iCurrentPressedRawVirtualKey)
    {
      iPressOccurredInsideCurrentVirtualKey = 
        aPointerEvent.iType == TPointerEvent::EButton1Down;
    }
    iCurrentPressedRawVirtualKey = rawVirtKey;
  }

  if (iCurrentPressedRawVirtualKey == ERawVirtKeyHide &&
      iPressOccurredInsideCurrentVirtualKey)
  {
    iVirtualKeyboardVisible = EFalse;
    iGameWindow.ClearScreen();
  }

  return insideVirtualKeyboard;
}
#endif

void
CGameController::KeyDown(TInt aScanCode)
{
  for (TInt i = 0; i < MAX_SIMULTANEOUS_KEYPRESSES; i++)
  {
    if (iScanCodes[i] == aScanCode)
    {
      // It seems we miss a key up event sometimes
      break;
    }
    else if (iScanCodes[i] == 0)
    {
      iScanCodes[i] = aScanCode;
      if (KeyNeedsAutoUp(aScanCode))
      {
        iAutoUpTimers[i] = 3;
      }
      iNoOfKeyPresses++;
      break;
    }
  }
}

void
CGameController::KeyUp(TInt aScanCode)
{
  for (TInt i=0 ; i<MAX_SIMULTANEOUS_KEYPRESSES; i++)
  {
    if (iScanCodes[i] == aScanCode)
    {
      iScanCodes[i] = 0;
      iNoOfKeyPresses--;
      break;
    }
  }
}

TBool
CGameController::IsKeyPressed(TInt aScanCode)
{
  TInt presses = 0;
  
  for (TInt i = 0; 
       i < MAX_SIMULTANEOUS_KEYPRESSES && presses < iNoOfKeyPresses; i++)
  {
    if (iScanCodes[i] != 0)
    {
      presses++;
      if (iScanCodes[i] == aScanCode)
      {
        if (KeyNeedsAutoUp(aScanCode))
        {
          iAutoUpTimers[i]--;
          if (iAutoUpTimers[i] == 0)
          {
            iScanCodes[i] = 0;
            iNoOfKeyPresses--;
          }
        }
        return ETrue;
      }
    }
  }

#if defined __UIQ__ || defined __NOKIA7710__
  if (iVirtualKeyboardVisible &&
      iCurrentPressedRawVirtualKey != ERawVirtKeyNone &&
      iPressOccurredInsideVirtualKeyboard)
  {
    switch (aScanCode)
    {
      case EMameVirtKeyUp:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyN ||
          iCurrentPressedRawVirtualKey == ERawVirtKeyNW ||
          iCurrentPressedRawVirtualKey == ERawVirtKeyNE;
      case EMameVirtKeyDown:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyS ||
          iCurrentPressedRawVirtualKey == ERawVirtKeySW ||
          iCurrentPressedRawVirtualKey == ERawVirtKeySE;
      case EMameVirtKeyLeft:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyW ||
          iCurrentPressedRawVirtualKey == ERawVirtKeyNW ||
          iCurrentPressedRawVirtualKey == ERawVirtKeySW;
      case EMameVirtKeyRight:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyE ||
          iCurrentPressedRawVirtualKey == ERawVirtKeyNE ||
          iCurrentPressedRawVirtualKey == ERawVirtKeySE;
      case EMameVirtKeyCenter:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyCenter;
      case EMameVirtKeyFire:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyFire &&
          iPressOccurredInsideCurrentVirtualKey;
      case EMameVirtKeyConf:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyConf &&
          iPressOccurredInsideCurrentVirtualKey;
      case EMameVirtKeySel:
        return iCurrentPressedRawVirtualKey == ERawVirtKeySel &&
          iPressOccurredInsideCurrentVirtualKey;
      case EMameVirtKeyEsc:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyEsc &&
          iPressOccurredInsideCurrentVirtualKey;
      case EMameVirtKeyCoin:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyCoin &&
          iPressOccurredInsideCurrentVirtualKey;
      case EMameVirtKeyStart:
        return iCurrentPressedRawVirtualKey == ERawVirtKeyStart &&
          iPressOccurredInsideCurrentVirtualKey;
    }
  }
#endif

  return EFalse;
}

#if defined __UIQ__ || defined __NOKIA7710__
void
CGameController::GetJoyStatus(TInt &x, TInt &y, TBool &pressed)
{
  const TInt springForce = 2;

  if (!iPressed)
  {
    if (iX > 0)
    {
      iX -= springForce;
      if (iX < 0)
      {
        iX = 0;
      }
    }
    if (iX < 0)
    {
      iX += springForce;
      if (iX > 0)
      {
        iX = 0;
      }
    }
    if (iY > 0)
    {
      iY -= springForce;
      if (iY < 0)
      {
        iY = 0;
      }
    }
    if (iY < 0)
    {
      iY += springForce;
      if (iY > 0)
      {
        iY = 0;
      }
    }
  }

  x = iX;
  y = iY;
  pressed = iPressed;
}
#endif

TBool
CGameController::KeyNeedsAutoUp(TInt aScanCode)
{
#ifdef __UIQ__
  return aScanCode == EStdKeyDeviceA ||
    aScanCode == EStdKeyApplication0 ||
    aScanCode == EStdKeyApplication1 ||
    aScanCode == EStdKeyYes ||
    aScanCode == EStdKeyNo ||
    aScanCode == EStdKeyIncVolume ||
    aScanCode == EStdKeyDecVolume;
#else
  return false;
#endif
}

#if defined __UIQ__ || defined __NOKIA7710__
void
CGameController::DrawVirtualKeyboard()
{
  if (iVirtualKeyboardVisible)
  {
    TUint16 *src = (TUint16 *)iVirtualKeyboardBitmap->DataAddress();

    TInt rows = iVirtualKeyboardSize.iHeight;
    TInt cols = iVirtualKeyboardSize.iWidth;
    for (int r = 0; r < rows; r++)
    {
      TUint16 *dst = 
        (TUint16 *)iGameWindow.GameScreenDevice()->
        ScanLineAddress(iVirtualKeyboardPosition.iY + r);
      for (int c = 0; c < cols; c++)
      {
        *dst++ = *src++;
      }
    }
  }
}
#endif

/**********************************************
 *
 * CDirectBitmapDrawer
 *
 **********************************************/

CDirectBitmapDrawer::CDirectBitmapDrawer(CGameWindow &aGameWindow)

  : CGameScreenDrawer(aGameWindow),
    iGameOrigin(0, 0),
    iNeedsPaning(EFalse)
{
}

void CDirectBitmapDrawer::DrawFrame(const struct osd_bitmap *bitmap)
{
  DirectDrawMameDisplay(bitmap);
  if (iNeedsPaning)
  {
    CheckForPaning();
  }
}

void
CDirectBitmapDrawer::OsdSetVisibleArea(TInt &min_x,
                                       TInt &max_x,
                                       TInt &min_y,
                                       TInt &max_y)
{
  iVisibleArea = TRect(min_x, min_y, max_x + 1, max_y + 1);
  UpdateVisibleArea(ETrue);
}

void
CDirectBitmapDrawer::UpdateVisibleArea(TBool aOsdVisibleAreaChanged)
{
 TSize screenSize = iGameWindow.GameScreenDevice()->Size();

 TBool needsPaningX = ETrue;
 TInt diffX = screenSize.iWidth - (iVisibleArea.iBr.iX - iVisibleArea.iTl.iX);
 if (diffX >= 0)
 {
   iGameOrigin.iX = diffX / 2 - iVisibleArea.iTl.iX;
   needsPaningX = EFalse;
 }
 else
 {
   if (aOsdVisibleAreaChanged)
   {
     iGameOrigin.iX = diffX / 2 - iVisibleArea.iTl.iX; // center
   }
   else if (iGameOrigin.iX < screenSize.iWidth - iVisibleArea.iBr.iX)
   {
     iGameOrigin.iX = screenSize.iWidth - iVisibleArea.iBr.iX;
   }
   else if (iGameOrigin.iX > -iVisibleArea.iTl.iX)
   {
     iGameOrigin.iX = -iVisibleArea.iTl.iX;
   }
 }

 TBool needsPaningY = ETrue;
 TInt diffY = screenSize.iHeight - (iVisibleArea.iBr.iY - iVisibleArea.iTl.iY);
 if (diffY > 0)
 {
   iGameOrigin.iY = diffY / 2 - iVisibleArea.iTl.iY;
   needsPaningY = EFalse;
 }
 else
 {
   if (aOsdVisibleAreaChanged)
   {
     iGameOrigin.iY = diffY / 2 - iVisibleArea.iTl.iY; // center
   }
   else if (iGameOrigin.iY < screenSize.iHeight - iVisibleArea.iBr.iY)
   {
     iGameOrigin.iY = screenSize.iHeight - iVisibleArea.iBr.iY;
   }
   else if (iGameOrigin.iY > -iVisibleArea.iTl.iY)
   {
     iGameOrigin.iY = -iVisibleArea.iTl.iY;
   }
 }

 iNeedsPaning = needsPaningX || needsPaningY;

 TRect screenInGameCoords(-iGameOrigin.iX,
                          -iGameOrigin.iY,
                          -iGameOrigin.iX + screenSize.iWidth,
                          -iGameOrigin.iY + screenSize.iHeight);
 iVisibleArea = TRect(iVisibleArea.iTl.iX, 
                      iVisibleArea.iTl.iY,
                      iVisibleArea.iBr.iX,
                      iVisibleArea.iBr.iY);
 TRect visibleAreaOnScreen(iVisibleArea);

 visibleAreaOnScreen.Intersection(screenInGameCoords);
 set_ui_visarea(visibleAreaOnScreen.iTl.iX, 
                visibleAreaOnScreen.iTl.iY,
                visibleAreaOnScreen.iBr.iX - 1,
                visibleAreaOnScreen.iBr.iY - 1);
 
 iUsedScreenArea = visibleAreaOnScreen;
 iUsedScreenArea.Move(TPoint(iGameOrigin));
}

void
CDirectBitmapDrawer::CheckForPaning()
{
  const TInt panStep = 5;
 
  TPoint delta(0, 0);
  if (input_ui_pressed_repeat(IPT_UI_PAN_RIGHT, 1))
  {
    delta.iX = panStep;
  }
  else if (input_ui_pressed_repeat(IPT_UI_PAN_LEFT, 1))
  {
    delta.iX = -panStep;
  }
  else if (input_ui_pressed_repeat(IPT_UI_PAN_UP, 1))
  {
    delta.iY = -panStep;
  }
  else if (input_ui_pressed_repeat(IPT_UI_PAN_DOWN, 1))
  {
    delta.iY = panStep;
  }
  if (delta.iX || delta.iY)
  {
    iGameOrigin -= delta;
    UpdateVisibleArea(EFalse);
  }
}

void
CDirectBitmapDrawer::DirectDrawMameDisplay(
  const struct osd_bitmap *bitmap)
{
  TRect rect(iVisibleArea);
  rect.Move(iGameOrigin);
  rect.Intersection(TRect(TPoint(0, 0),
                          iGameWindow.GameScreenDevice()->Size()));
  TRect sourceRect(rect);
  sourceRect.Move(-iGameOrigin);

  switch (iGameWindow.GameScreenDevice()->DisplayMode())
  {
    case EColor4K:
    case EColor64K:
      switch (bitmap->depth)
      {
        case 8:
          DirectDraw8BitMameDisplayTo12or16Bpp(bitmap,
                                               sourceRect.iTl,
                                               rect.iTl,
                                               rect.Size());
          break;
        default:
          ASSERT(EFalse);
      }
      break;
#ifdef __HAS_COLOR_16MU__
    case EColor16MU:
      switch (bitmap->depth)
      {
        case 8:
          DirectDraw8BitMameDisplayTo24Bpp(bitmap,
                                           sourceRect.iTl,
                                           rect.iTl,
                                           rect.Size());
          break;
        default:
          ASSERT(EFalse);
      }
      break;
#endif
    default:
      ASSERT(EFalse);
      break;
  }
}

void
CDirectBitmapDrawer::DirectDraw8BitMameDisplayTo12or16Bpp(
  const struct osd_bitmap *bitmap,
  TPoint srcPoint,
  TPoint dstPoint,
  TSize size)
{
  int addPerRow = bitmap->line[1] - bitmap->line[0];
  TUint8 *src_line = (TUint8 *)bitmap->line[srcPoint.iY] + srcPoint.iX;

  TUint16 *dst_line =
    (TUint16 *)TGameBitmapUtil::PixelAddress(iGameWindow.GameScreenDevice(), 
                                             dstPoint);
  TInt dstLineDiff = iGameWindow.GameScreenDevice()->BytesPerLine() / 2;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();

  TInt width = size.iWidth;
  for (TInt h = size.iHeight; h > 0; h--)
  {
    TUint16 *dst = dst_line;
    const TUint8 *src = src_line;
    for (TInt w = width; w > 0; w--)
    {
      TUint color = pal[*src];
      src++;
      *dst++ = (TUint16)color;
    }
    dst_line += dstLineDiff;
    src_line += addPerRow;
  }
}

void
CDirectBitmapDrawer::DirectDraw8BitMameDisplayTo24Bpp(
  const struct osd_bitmap *bitmap,
  TPoint srcPoint,
  TPoint dstPoint,
  TSize size)
{
  int addPerRow = bitmap->line[1] - bitmap->line[0];
  TUint8 *src_line = (TUint8 *)bitmap->line[srcPoint.iY] + srcPoint.iX;

  TUint32 *dst_line =
    (TUint32 *)TGameBitmapUtil::PixelAddress(iGameWindow.GameScreenDevice(), 
                                             dstPoint);
  TInt dstLineDiff = iGameWindow.GameScreenDevice()->BytesPerLine() / 4;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();

  TInt width = size.iWidth;
  for (TInt h = size.iHeight; h > 0; h--)
  {
    TUint32 *dst = dst_line;
    const TUint8 *src = src_line;
    for (TInt w = width; w > 0; w--)
    {
      TUint color = pal[*src];
      src++;
      *dst++ = color;
    }
    dst_line += dstLineDiff;
    src_line += addPerRow;
  }
}

/********************************************
 *
 * CScaledBitmapDrawer implementation
 *
 ********************************************/

CScaledBitmapDrawer::CScaledBitmapDrawer(CGameWindow &aGameWindow)
  : CGameScreenDrawer(aGameWindow)
{
}

void 
CScaledBitmapDrawer::DrawFrame(const struct osd_bitmap *bitmap)
{
  DrawScaledMameDisplay(bitmap);
}

void
CScaledBitmapDrawer::OsdSetVisibleArea(TInt &min_x,
                                       TInt &max_x,
                                       TInt &min_y,
                                       TInt &max_y)
{
  iVisibleArea = TRect(min_x, min_y, max_x + 1, max_y + 1);
  ComputeScaling();
  set_ui_visarea(min_x, min_y, max_x, max_y);
}

void
CScaledBitmapDrawer::ComputeScaling()
{
  TSize dstSize(iGameWindow.GameScreenDevice()->Size());

  int displayWidth = iVisibleArea.iBr.iX - iVisibleArea.iTl.iX;
  int displayHeight = iVisibleArea.iBr.iY - iVisibleArea.iTl.iY;

  iAddX = 1;
  iModuloX = 1;
  iAddY = 1;
  iModuloY = 1;
  TSize scaledSize = TSize(displayWidth, displayHeight);

  if (dstSize.iWidth < displayWidth)
  {
    iAddX = dstSize.iWidth;
    iModuloX = displayWidth;
    scaledSize.iWidth = dstSize.iWidth;
  }
  if (dstSize.iHeight < displayHeight)
  {
    iAddY = dstSize.iHeight;
    iModuloY = displayHeight;
    scaledSize.iHeight = dstSize.iHeight;
  }
  if (iKeepAspectRatio &&
      (dstSize.iWidth < displayWidth || dstSize.iHeight < displayHeight))
  {
    if (displayHeight * dstSize.iWidth < dstSize.iHeight * displayWidth)
    {
      // scale is determined by horizontal scale
      iAddY = iAddX;
      iModuloY = iModuloX;
      scaledSize.iHeight =
        (displayHeight * dstSize.iWidth + displayWidth - 1) / displayWidth;
    }
    else
    {
      // scale is determined by vertical scale
      iAddX = iAddY;
      iModuloX = iModuloY;
      scaledSize.iWidth =
        (displayWidth * dstSize.iHeight + displayHeight - 1) / displayHeight;
    }
  }

  TPoint point((dstSize.iWidth - scaledSize.iWidth) / 2,
               (dstSize.iHeight - scaledSize.iHeight) / 2);

  iUsedScreenArea = TRect(point, scaledSize);
}

void
CScaledBitmapDrawer::DrawScaledMameDisplay(
  const struct osd_bitmap *bitmap)
{
  switch (iGameWindow.GameScreenDevice()->DisplayMode())
  {
    case EColor4K:
    case EColor64K:
      switch (bitmap->depth)
      {
        case 8:
          DrawScaled8BitMameDisplayTo12or16Bpp(bitmap);
          break;
        default:
          ASSERT(EFalse);
      }
      break;
#ifdef __HAS_COLOR_16MU__
    case EColor16MU:
      switch (bitmap->depth)
      {
        case 8:
          DrawScaled8BitMameDisplayTo24Bpp(bitmap);
          break;
        default:
          ASSERT(EFalse);
      }
      break;
#endif
    default:
      ASSERT(EFalse);
      break;
  }
}

/**********************************************
 *
 * CFastScaledBitmapDrawer
 *
 **********************************************/

CFastScaledBitmapDrawer::CFastScaledBitmapDrawer(CGameWindow &aGameWindow)
  : CScaledBitmapDrawer(aGameWindow)
{
}

void
CFastScaledBitmapDrawer::DrawScaled8BitMameDisplayTo12or16Bpp(
  const struct osd_bitmap *bitmap)
{  
  TInt mask = 0; // Clears lsb for each color component
  switch (iGameWindow.GameScreenDevice()->DisplayMode())
  {
    case EColor4K:
      mask = 0x0eee; // 0000 bbbb gggg rrrr : 0000 1110 1110 1110
      break;
    case EColor64K:
      mask = 0xf7de; // bbbb bggg gggr rrrr : 1111 0111 1101 1110
      break;
    default:
      ASSERT(EFalse);
      break;
  }

  TUint8 *src = (TUint8 *)bitmap->line[0];
  int srcAddPerRow = bitmap->line[1] - bitmap->line[0];

  int dstAddPerRow = iGameWindow.GameScreenDevice()->BytesPerLine() / 2;

  TInt srcLines = iVisibleArea.Height();
  TInt srcColumns = iVisibleArea.Width();
  TInt accumulator = 0;
  TBool merge = EFalse;

  src += iVisibleArea.iTl.iY * srcAddPerRow + iVisibleArea.iTl.iX;
  TUint16 *dst =
    (TUint16 *)TGameBitmapUtil::PixelAddress(iGameWindow.GameScreenDevice(), 
                                             iUsedScreenArea.iTl);

  for (TInt h = 0; h < srcLines; h++)
  {
    if (merge)
    {
      DrawScaled8BitMameLineTo12or16BppWithMerge(src,
                                                 iAddX,
                                                 iModuloX,
                                                 dst,
                                                 srcColumns,
                                                 mask);
    }
    else
    {
      DrawScaled8BitMameLineTo12or16Bpp(src,
                                        iAddX,
                                        iModuloX,
                                        dst,
                                        srcColumns,
                                        mask);
    }
    merge = ETrue;
    accumulator += iAddY;
    if (accumulator >= iModuloY || h == srcLines - 1)
    {
      dst += dstAddPerRow;
      merge = EFalse;
      accumulator -= iModuloY;
    }
    src += srcAddPerRow;
  }
}

void
CFastScaledBitmapDrawer::DrawScaled8BitMameDisplayTo24Bpp(
  const struct osd_bitmap *bitmap)
{  
  TInt mask = 0; // Clears lsb for each color component
  switch (iGameWindow.GameScreenDevice()->DisplayMode())
  {
#ifdef __HAS_COLOR_16MU__
    case EColor16MU:
      mask = 0xfefefe; // bbbb bbbb gggg gggg rrrr rrrr
      break;
#endif
    default:
      ASSERT(EFalse);
      break;
  }

  TUint8 *src = (TUint8 *)bitmap->line[0];
  int srcAddPerRow = bitmap->line[1] - bitmap->line[0];

  int dstAddPerRow = iGameWindow.GameScreenDevice()->BytesPerLine() / 4;

  TInt srcLines = iVisibleArea.Height();
  TInt srcColumns = iVisibleArea.Width();
  TInt accumulator = 0;
  TBool merge = EFalse;

  src += iVisibleArea.iTl.iY * srcAddPerRow + iVisibleArea.iTl.iX;
  TUint32 *dst =
    (TUint32 *)TGameBitmapUtil::PixelAddress(iGameWindow.GameScreenDevice(), 
                                             iUsedScreenArea.iTl);
  for (TInt h = 0; h < srcLines; h++)
  {
    if (merge)
    {
      DrawScaled8BitMameLineTo24BppWithMerge(src,
                                             iAddX,
                                             iModuloX,
                                             dst,
                                             srcColumns,
                                             mask);
    }
    else
    {
      DrawScaled8BitMameLineTo24Bpp(src,
                                    iAddX,
                                    iModuloX,
                                    dst,
                                    srcColumns,
                                    mask);
    }
    merge = ETrue;
    accumulator += iAddY;
    if (accumulator >= iModuloY || h == srcLines - 1)
    {
      dst += dstAddPerRow;
      merge = EFalse;
      accumulator -= iModuloY;
    }
    src += srcAddPerRow;
  }
}

#define COMPUTE_RGB_AVERAGE(a, b, mask) ((((a)&(mask))>>1)+(((b)&(mask))>>1)+((a)&(~(mask))))

void
CFastScaledBitmapDrawer::DrawScaled8BitMameLineTo12or16Bpp(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint16 *dst,
  int visibleColumns,
  TInt mask)
{

  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_RGB_AVERAGE(pixel, srcPixel, mask);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}

void
CFastScaledBitmapDrawer::DrawScaled8BitMameLineTo12or16BppWithMerge(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint16 *dst,
  int visibleColumns,
  TInt mask)
{

  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_RGB_AVERAGE(pixel, srcPixel, mask);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      pixel = COMPUTE_RGB_AVERAGE(*dst, pixel, mask); // merge
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}

void
CFastScaledBitmapDrawer::DrawScaled8BitMameLineTo24Bpp(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint32 *dst,
  int visibleColumns,
  TInt mask)
{

  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_RGB_AVERAGE(pixel, srcPixel, mask);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}

void
CFastScaledBitmapDrawer::DrawScaled8BitMameLineTo24BppWithMerge(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint32 *dst,
  int visibleColumns,
  TInt mask)
{

  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_RGB_AVERAGE(pixel, srcPixel, mask);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      pixel = COMPUTE_RGB_AVERAGE(*dst, pixel, mask); // merge
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}


/**********************************************
 *
 * CGammaCorrectScaledBitmapDrawer
 *
 **********************************************/

CGammaCorrectScaledBitmapDrawer::CGammaCorrectScaledBitmapDrawer(
  CGameWindow &aGameWindow)
  : CScaledBitmapDrawer(aGameWindow)
{
}

CGammaCorrectScaledBitmapDrawer *
CGammaCorrectScaledBitmapDrawer::NewL(CGameWindow &aGameWindow)
{
  CGammaCorrectScaledBitmapDrawer *self = 
    new CGammaCorrectScaledBitmapDrawer(aGameWindow);
  CleanupStack::PushL(self);
  TInt bitsPerColor = 0;
  switch (aGameWindow.GameScreenDevice()->DisplayMode())
  {
    case EColor4K:
      bitsPerColor = 4;
      break;
    case EColor64K:
      bitsPerColor = 5; // We ignore the fact that green has 6 bits
      break;
#ifdef __HAS_COLOR_16MU__
    case EColor16MU:
      bitsPerColor = 8;
      break;
#endif
    default:
      ASSERT(EFalse);
  }
  self->ComputeComponentMixTableL(bitsPerColor);
  CleanupStack::Pop();
  return self;
}

CGammaCorrectScaledBitmapDrawer::~CGammaCorrectScaledBitmapDrawer()
{
  User::FreeZ((TAny *&)iComponentMixTable);
}

void
CGammaCorrectScaledBitmapDrawer::DrawScaled8BitMameDisplayTo12or16Bpp(
  const struct osd_bitmap *bitmap)
{  
  TDisplayMode mode = iGameWindow.GameScreenDevice()->DisplayMode();

  TUint8 *src = (TUint8 *)bitmap->line[0];
  int srcAddPerRow = bitmap->line[1] - bitmap->line[0];

  int dstAddPerRow = iGameWindow.GameScreenDevice()->BytesPerLine() / 2;

  TInt srcLines = iVisibleArea.Height();
  TInt srcColumns = iVisibleArea.Width();
  TInt accumulator = 0;
  TBool merge = EFalse;

  src += iVisibleArea.iTl.iY * srcAddPerRow + iVisibleArea.iTl.iX;
  TUint16 *dst =
    (TUint16 *)TGameBitmapUtil::PixelAddress(iGameWindow.GameScreenDevice(), 
                                             iUsedScreenArea.iTl);
  for (TInt h = 0; h < srcLines; h++)
  {
    switch (mode)
    {
      case EColor4K:
        if (merge)
        {
          DrawGammaCorrectScaled8BitMameLineTo12BppWithMerge(src,
                                                             iAddX,
                                                             iModuloX,
                                                             dst,
                                                             srcColumns);
        }
        else
        {
          DrawGammaCorrectScaled8BitMameLineTo12Bpp(src,
                                                    iAddX,
                                                    iModuloX,
                                                    dst,
                                                    srcColumns);
        }
        break;
      case EColor64K:
        if (merge)
        {
          DrawGammaCorrectScaled8BitMameLineTo16BppWithMerge(src,
                                                             iAddX,
                                                             iModuloX,
                                                             dst,
                                                             srcColumns);
        }
        else
        {
          DrawGammaCorrectScaled8BitMameLineTo16Bpp(src,
                                                    iAddX,
                                                    iModuloX,
                                                    dst,
                                                    srcColumns);
        }
        break;
      default:
        ASSERT(EFalse);
    }
    merge = ETrue;
    accumulator += iAddY;
    if (accumulator >= iModuloY || h == srcLines - 1)
    {
      dst += dstAddPerRow;
      merge = EFalse;
      accumulator -= iModuloY;
    }
    src += srcAddPerRow;
  }
}


void
CGammaCorrectScaledBitmapDrawer::DrawScaled8BitMameDisplayTo24Bpp(
  const struct osd_bitmap *bitmap)
{  
  TDisplayMode mode = iGameWindow.GameScreenDevice()->DisplayMode();

  TUint8 *src = (TUint8 *)bitmap->line[0];
  int srcAddPerRow = bitmap->line[1] - bitmap->line[0];

  int dstAddPerRow = iGameWindow.GameScreenDevice()->BytesPerLine() / 4;

  TInt srcLines = iVisibleArea.Height();
  TInt srcColumns = iVisibleArea.Width();
  TInt accumulator = 0;
  TBool merge = EFalse;

  src += iVisibleArea.iTl.iY * srcAddPerRow + iVisibleArea.iTl.iX;
  TUint32 *dst =
    (TUint32 *)TGameBitmapUtil::PixelAddress(iGameWindow.GameScreenDevice(), 
                                             iUsedScreenArea.iTl);
  for (TInt h = 0; h < srcLines; h++)
  {
    switch (mode)
    {
#ifdef __HAS_COLOR_16MU__
      case EColor16MU:
        if (merge)
        {
          DrawGammaCorrectScaled8BitMameLineTo24BppWithMerge(src,
                                                             iAddX,
                                                             iModuloX,
                                                             dst,
                                                             srcColumns);
        }
        else
        {
          DrawGammaCorrectScaled8BitMameLineTo24Bpp(src,
                                                    iAddX,
                                                    iModuloX,
                                                    dst,
                                                    srcColumns);
        }
        break;
#endif
      default:
        ASSERT(EFalse);
    }
    merge = ETrue;
    accumulator += iAddY;
    if (accumulator >= iModuloY || h == srcLines - 1)
    {
      dst += dstAddPerRow;
      merge = EFalse;
      accumulator -= iModuloY;
    }
    src += srcAddPerRow;
  }
}

void
CGammaCorrectScaledBitmapDrawer::ComputeComponentMixTableL(TInt aBits)
{
  const TReal gamma = 2.1;
  const TReal gammaInv = 1/gamma;

  iComponentMixTable = 
    (TInt *)User::AllocL(sizeof (TInt) * (1 << (2 * aBits)));

  for (TInt x = 0; x < (1 << aBits); x++)
  {
    for (TInt y = 0; y < x; y++)
    {
      TReal realX = x;
      TReal realY = y;

      TReal displayedX;
      Math::Pow(displayedX, realX, gamma);
      TReal displayedY;
      Math::Pow(displayedY, realY, gamma);

      TReal displayedAverage = (displayedX + displayedY) / 2;
      TReal realZ;
      Math::Pow(realZ, displayedAverage, gammaInv);
      
      TInt averageLo = (TInt)(realZ + 0.25);
      TInt averageHi = (TInt)(realZ + 0.75);
      ASSERT(averageHi < (2 << aBits));

      iComponentMixTable[(x << aBits) | y] = averageLo;
      iComponentMixTable[(y << aBits) | x] = averageHi;
    }
    iComponentMixTable[(x << aBits) | x] = x;
  }
}

#define COMPUTE_12BIT_GAMMA_CORRECT_RGB_AVERAGE(a, b) \
  ((iComponentMixTable[((a & 0xf00) >> 4) | ((b & 0xf00) >> 8)] << 8) | \
   (iComponentMixTable[((a & 0x0f0) >> 0) | ((b & 0x0f0) >> 4)] << 4) | \
    iComponentMixTable[((a & 0x00f) << 4) | ((b & 0x00f) >> 0)])

void
CGammaCorrectScaledBitmapDrawer::DrawGammaCorrectScaled8BitMameLineTo12Bpp(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint16 *dst,
  int visibleColumns)
{
  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_12BIT_GAMMA_CORRECT_RGB_AVERAGE(pixel, srcPixel);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}

void
CGammaCorrectScaledBitmapDrawer::DrawGammaCorrectScaled8BitMameLineTo12BppWithMerge(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint16 *dst,
  int visibleColumns)
{

  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_12BIT_GAMMA_CORRECT_RGB_AVERAGE(pixel, srcPixel);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      pixel = COMPUTE_12BIT_GAMMA_CORRECT_RGB_AVERAGE(*dst, pixel);
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}

// bbbb bggg gggr rrrr - ignore lsb of green : bbbb bggg gg0r rrrr
#define COMPUTE_16BIT_GAMMA_CORRECT_RGB_AVERAGE(a, b) \
  ((iComponentMixTable[((a & 0xf800) >> 6) | ((b & 0xf800) >> 11)] << 11) | \
   (iComponentMixTable[((a & 0x07c0) >> 1) | ((b & 0x07c0) >> 6)] << 6) | \
    iComponentMixTable[((a & 0x001f) << 5) | ((b & 0x001f) >> 0)])

void
CGammaCorrectScaledBitmapDrawer::DrawGammaCorrectScaled8BitMameLineTo16Bpp(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint16 *dst,
  int visibleColumns)
{
  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_16BIT_GAMMA_CORRECT_RGB_AVERAGE(pixel, srcPixel);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}

void
CGammaCorrectScaledBitmapDrawer::DrawGammaCorrectScaled8BitMameLineTo16BppWithMerge(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint16 *dst,
  int visibleColumns)
{

  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_16BIT_GAMMA_CORRECT_RGB_AVERAGE(pixel, srcPixel);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      pixel = COMPUTE_16BIT_GAMMA_CORRECT_RGB_AVERAGE(*dst, pixel);
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}

#define COMPUTE_24BIT_GAMMA_CORRECT_RGB_AVERAGE(a, b) \
((iComponentMixTable[((a & 0xff0000) >> 8) | ((b & 0xff0000) >> 16)] << 16) | \
 (iComponentMixTable[((a & 0x00ff00) >> 0) | ((b & 0x00ff00) >> 8)] << 8) | \
  iComponentMixTable[((a & 0x0000ff) << 8) | ((b & 0x0000ff) >> 0)])

void
CGammaCorrectScaledBitmapDrawer::DrawGammaCorrectScaled8BitMameLineTo24Bpp(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint32 *dst,
  int visibleColumns)
{
  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_24BIT_GAMMA_CORRECT_RGB_AVERAGE(pixel, srcPixel);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}

void
CGammaCorrectScaledBitmapDrawer::DrawGammaCorrectScaled8BitMameLineTo24BppWithMerge(
  TUint8 *src_start,
  int add,
  int modulo,
  TUint32 *dst,
  int visibleColumns)
{

  TInt accumulator = 0;

  TUint *pal = iGameWindow.GameScreenDevice()->Palette()->PaletteBuffer();
  TUint8 *src = src_start;

  TInt pixel = 0;
  for (TInt w = visibleColumns - 1; w >= 0; w--)
  {
    TUint srcPixel = pal[*src];
    src++;
    if (accumulator >= add)
    {
      srcPixel = COMPUTE_24BIT_GAMMA_CORRECT_RGB_AVERAGE(pixel, srcPixel);
    }
    pixel = srcPixel;
    
    accumulator += add;
    if (accumulator >= modulo || w == 0)
    {
      pixel = COMPUTE_24BIT_GAMMA_CORRECT_RGB_AVERAGE(*dst, pixel);
      *dst = pixel;
      dst++;
      accumulator -= modulo;
    }
  }
}

/******************************************
 *
 * CGameDirectScreenAccess
 *
 ******************************************/

CGameDirectScreenAccess::~CGameDirectScreenAccess()
{
  if (iDirectScreenAccess != NULL)
  {
    if (!gDirectScreenAccessAborted)
    {
      iDirectScreenAccess->Cancel();
    }
    delete iDirectScreenAccess;
    iDirectScreenAccess = NULL;
  }
  iDirectScreenUpdateRegion.Close();
}

CGameDirectScreenAccess *CGameDirectScreenAccess::NewL(
  CGameWindow& aGameWindow)
{
  CGameDirectScreenAccess *self =
    new(ELeave)CGameDirectScreenAccess(aGameWindow);
  CleanupStack::PushL(self);
  self->ConstructL();
  CleanupStack::Pop();
  return self;
}

CGameDirectScreenAccess::CGameDirectScreenAccess(CGameWindow& aGameWindow) 
  : iGameWindow(aGameWindow)
{
}

void CGameDirectScreenAccess::ConstructL()
{
  iDirectScreenAccess = CDirectScreenAccess::NewL(iGameWindow.iWsSession,
                                                  *iGameWindow.iWsScreen,
                                                  iGameWindow.iWsWindow,
                                                  *this);

  iDirectScreenUpdateRegion.AddRect(TRect(iGameWindow.iWsWindow.Position(),
                                          iGameWindow.iWsWindow.Size()));

  /*
   * This flag could probably better be inverted and named, e.g.,
   * gHasDirectScreenAccess.
   */
  gDirectScreenAccessAborted = ETrue;

  TryToStartDrawing(ETrue);
}

void
CGameDirectScreenAccess::TryToStartDrawing(TBool aResumeFromFrontEnd)
{
  if (aResumeFromFrontEnd)
  {
    iGameWindow.Show();
  }
  iGameWindow.iWsSession.Flush();

  TRAPD(err, iDirectScreenAccess->StartL());
  if (err)
  {
    gDirectScreenAccessAborted = ETrue;
    return;
  }
  gDirectScreenAccessAborted = EFalse;

  RRegion *region = iDirectScreenAccess->DrawingRegion();

  TBool	isFullScreen = region->Count() == 1 && 
    (*region)[0] == TRect(iGameWindow.iWsWindow.Position(), 
                          iGameWindow.iWsWindow.Size());
  if (isFullScreen)
  {
    iGameWindow.ClearScreen();
  }
  else
  {
    iDirectScreenAccess->Cancel();
    gDirectScreenAccessAborted = ETrue;
  }
}

void 
CGameDirectScreenAccess::AbortNow(
  RDirectScreenAccess::TTerminationReasons)
{
  iDirectScreenAccess->Cancel();
  gDirectScreenAccessAborted = ETrue;
  iGameWindow.Hide();
}

void 
CGameDirectScreenAccess::Restart(
  RDirectScreenAccess::TTerminationReasons)
{
  if (gDirectScreenAccessAborted)
  {
    TryToStartDrawing(EFalse);
  }
}

void
CGameDirectScreenAccess::Update()
{
  iDirectScreenAccess->ScreenDevice()->Update(iDirectScreenUpdateRegion);
}
