/*******************************************************************
 *
 *	File:		Video.h
 *
 *	Author:		Peter van Sebille (peter@yipton.net)
 *
 *      Modified 2003 for Series 60 by Staffan Ulfberg
 *      (staffan@ulfberg.se).  Lots of P800 specific things, and stuff
 *      not used by EMame, was removed.
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#ifndef __VIDEO_H
#define __VIDEO_H


#include <e32base.h>
#include <w32std.h>

#include "target.h"

class CGamePalette : public CBase
{
  public:
    ~CGamePalette();
    static CGamePalette* NewL(TDisplayMode aDisplayMode);
    
    TUint *PaletteBuffer() const 
    {
      return (TUint *)iPalette;
    }

    TInt AllocateEntry(const TRgb& aColor, TBool aTryToSaveSpace = ETrue);
    TDisplayMode DisplayMode() const {return iDisplayMode;}

    void ModifyPen(int pen, 
                   unsigned char red,
                   unsigned char green,
                   unsigned char blue);
    void GetPen(int pen,
                unsigned char *red,
                unsigned char *green,
                unsigned char *blue);
    TInt GetPaletteMaxIndex();

protected:
    CGamePalette(TDisplayMode aDisplayMode);
    void ConstructL();
    
    TUint iPalette[256];
    TDisplayMode iDisplayMode;
    TInt iPaletteLastIndexUsed;
    TInt iPaletteMaxIndex;
};

class CGameScreenDevice : public CBase
{
  public:
    ~CGameScreenDevice();
    
    // getters
    virtual TUint8 *ScanLineAddress(TInt aLine) const = 0;
    TInt BytesPerLine() const {return iBytesPerLine;}
    TSize Size() const {return iSize;}
    TDisplayMode DisplayMode() const 
    {
      return iDisplayMode;
    }

    // draw operations
    void Clear();

    CGamePalette *Palette() const 
    {
      return iPalette;
    }
    
    void SetPalette(CGamePalette* aGamePalette)
    {
      iPalette = aGamePalette;
    }

  protected:
    CGameScreenDevice(const TSize& aSize, TDisplayMode aDisplayMode);

    TInt iBytesPerLine;
    TSize iSize;
    TDisplayMode iDisplayMode;
    CGamePalette *iPalette;
};


class CFbsGameScreenDevice : public CGameScreenDevice
{
 public:
  ~CFbsGameScreenDevice();
  static CFbsGameScreenDevice* NewL(const TSize& aSize, TDisplayMode aDisplayMode);
  virtual TUint8 *ScanLineAddress(TInt aLine) const;
  
  CFbsBitmap *FbsBitmap(){return iBitmap;}
  
 protected:
  CFbsGameScreenDevice(const TSize& aSize, TDisplayMode aDisplayMode);
  void ConstructL();
  
  CFbsBitmap *iBitmap;
};



class CHwFbGameScreenDevice : public CGameScreenDevice
{
 public:
  static CHwFbGameScreenDevice * 
    NewL(const TSize& aWindowSize, TDisplayMode aDisplayMode);
  
  virtual TUint8 *ScanLineAddress(TInt aLine) const;
  
 protected:
  CHwFbGameScreenDevice(const TSize& aWindowSize, TDisplayMode aDisplayMode);
  void ConstructL(const TScreenInfoV01& aScreenInfo);

  TUint8 *iFrameBuffer;
};

/*****************************************
 *
 * TGameBitmapUtil
 *
 *****************************************/

class TGameBitmapUtil
{
public:
	static TInt			GetBpp(TDisplayMode aDisplayMode);
	static TInt			GetColorDepth(TDisplayMode aDisplayMode);
	static TUint8*		PixelAddress(const CGameScreenDevice* aSrc, const TPoint& aPoint);
	static TUint		GetColorValue(const TRgb& aColor, const TDisplayMode& aDisplayMode);
    static TRgb GetRgb(TUint aColor, const TDisplayMode& aDisplayMode);
};

class CGameWindow;
class CGameController;
class CGameDirectScreenAccess;

/**********************************************
 *
 * CGameScreenDrawer
 *
 **********************************************/

class CGameScreenDrawer
{
  public:
    CGameScreenDrawer(CGameWindow &aGameWindow)
      : iGameWindow(aGameWindow),
        iKeepAspectRatio(ETrue)
    {
    }

    virtual void DrawFrame(const struct osd_bitmap *bitmap) = 0;
    virtual void OsdSetVisibleArea(TInt &min_x, 
				   TInt &max_x, 
				   TInt &min_y, 
				   TInt &max_y) = 0;
    TRect GetUsedScreenArea()
    {
      return iUsedScreenArea;
    }

    void SetKeepAspectRatio(TInt aKeepAspectRatio)
    {
      iKeepAspectRatio = aKeepAspectRatio;
    }

  protected:
    CGameWindow &iGameWindow;
    TRect iVisibleArea;
    TRect iUsedScreenArea;
    TBool iKeepAspectRatio; // Only used for scaling
};

/**********************************************
 *
 * CDirectBitmapDrawer
 *
 **********************************************/

class CDirectBitmapDrawer : public CGameScreenDrawer
{
  public:
    CDirectBitmapDrawer(CGameWindow &aGameWindow);
    virtual void DrawFrame(const struct osd_bitmap *bitmap);
    virtual void OsdSetVisibleArea(TInt &min_x,
                                   TInt &max_x,
                                   TInt &min_y,
                                   TInt &max_y);

  private:
    void UpdateVisibleArea(TBool aOsdVisibleAreaChanged);
    void CheckForPaning();

    void DirectDrawMameDisplay(const struct osd_bitmap *bitmap);
    void DirectDraw8BitMameDisplayTo12or16Bpp(
      const struct osd_bitmap *bitmap,
      TPoint srcPoint,
      TPoint dstPoint,
      TSize size);
    void DirectDraw8BitMameDisplayTo24Bpp(
      const struct osd_bitmap *bitmap,
      TPoint srcPoint,
      TPoint dstPoint,
      TSize size);
    
    TPoint iGameOrigin;
    TBool iNeedsPaning;
};

/**********************************************
 *
 * CScaledBitmapDrawer
 *
 **********************************************/

class CScaledBitmapDrawer : public CGameScreenDrawer
{
public:
    CScaledBitmapDrawer(CGameWindow &aGameWindow);

    virtual void DrawFrame(const struct osd_bitmap *bitmap);

    virtual void OsdSetVisibleArea(TInt &min_x,
                                   TInt &max_x,
                                   TInt &min_y,
                                   TInt &max_y);

  protected:
    virtual void DrawScaled8BitMameDisplayTo12or16Bpp(
      const struct osd_bitmap *bitmap) = 0;
    virtual void DrawScaled8BitMameDisplayTo24Bpp(
      const struct osd_bitmap *bitmap) = 0;

  private:
    void DrawScaledMameDisplay(const struct osd_bitmap *bitmap);
    void ComputeScaling();

  protected:
    TInt iAddX;
    TInt iModuloX;
    TInt iAddY;
    TInt iModuloY;
};


/**********************************************
 *
 * CFastScaledBitmapDrawer
 *
 **********************************************/

class CFastScaledBitmapDrawer : public CScaledBitmapDrawer
{
  public:
    CFastScaledBitmapDrawer(CGameWindow &aGameWindow);

  protected:
    virtual void DrawScaled8BitMameDisplayTo12or16Bpp(
      const struct osd_bitmap *bitmap);
    virtual void DrawScaled8BitMameDisplayTo24Bpp(
      const struct osd_bitmap *bitmap);

  private:

    void DrawScaled8BitMameLineTo12or16Bpp(TUint8 *src_start,
                                           int add,
                                           int modulo,
                                           TUint16 *dst,
                                           int visibleColumns,
					   TInt mask);
    void DrawScaled8BitMameLineTo12or16BppWithMerge(TUint8 *src_start,
						    int add,
						    int modulo,
						    TUint16 *dst,
						    int visibleColumns,
						    TInt mask);
    void DrawScaled8BitMameLineTo24Bpp(TUint8 *src_start,
                                           int add,
                                           int modulo,
                                           TUint32 *dst,
                                           int visibleColumns,
					   TInt mask);
    void DrawScaled8BitMameLineTo24BppWithMerge(TUint8 *src_start,
						int add,
						int modulo,
						TUint32 *dst,
						int visibleColumns,
						TInt mask);
};

/**********************************************
 *
 * CGammaCorrectScaledBitmapDrawer
 *
 **********************************************/

class CGammaCorrectScaledBitmapDrawer : public CScaledBitmapDrawer
{
  public:
    static CGammaCorrectScaledBitmapDrawer *NewL(CGameWindow &aGameWindow);
    virtual ~CGammaCorrectScaledBitmapDrawer();

  protected:
    virtual void DrawScaled8BitMameDisplayTo12or16Bpp(
      const struct osd_bitmap *bitmap);
    virtual void DrawScaled8BitMameDisplayTo24Bpp(
      const struct osd_bitmap *bitmap);

  private:
    CGammaCorrectScaledBitmapDrawer(CGameWindow &aGameWindow);
    void ComputeComponentMixTableL(TInt aBits);

    void DrawGammaCorrectScaled8BitMameLineTo12Bpp(TUint8 *src_start,
						   int add,
						   int modulo,
						   TUint16 *dst,
						   int visibleColumns);
    
    void DrawGammaCorrectScaled8BitMameLineTo12BppWithMerge(
      TUint8 *src_start,
      int add,
      int modulo,
      TUint16 *dst,
      int visibleColumns);

    void DrawGammaCorrectScaled8BitMameLineTo16Bpp(TUint8 *src_start,
						   int add,
						   int modulo,
						   TUint16 *dst,
						   int visibleColumns);
    
    void DrawGammaCorrectScaled8BitMameLineTo16BppWithMerge(
      TUint8 *src_start,
      int add,
      int modulo,
      TUint16 *dst,
      int visibleColumns);

    void DrawGammaCorrectScaled8BitMameLineTo24Bpp(TUint8 *src_start,
						   int add,
						   int modulo,
						   TUint32 *dst,
						   int visibleColumns);
    
    void DrawGammaCorrectScaled8BitMameLineTo24BppWithMerge(
      TUint8 *src_start,
      int add,
      int modulo,
      TUint32 *dst,
      int visibleColumns);

    TInt *iComponentMixTable;
};

/********************************************
 *
 * CGameWindow
 *
 ********************************************/

const TInt KMaxCapturedKeys = 10;

class CGameWindow : public CBase
{
  public:
    ~CGameWindow();
    static CGameWindow* NewL(const TDesC& aGameName, 
			     TBool aVirtualKeyboardEnabled);

    CGameScreenDevice *GameScreenDevice()
    {
      return iGameScreenDevice;
    }

    void OsdSetVisibleArea(TInt &min_x, TInt &max_x, TInt &min_y, TInt &max_y)
    {
      iGameScreenDrawer->OsdSetVisibleArea(min_x, max_x, min_y, max_y);
      ClearScreen();
    }

    void SetKeepAspectRatio(TInt aKeepAspectRatio)
    {
      iGameScreenDrawer->SetKeepAspectRatio(aKeepAspectRatio);
      ClearScreen();
    }

    virtual TRect GetUsedScreenArea()
    {
      return iGameScreenDrawer->GetUsedScreenArea();
    }

    void GameFrameTick(struct osd_bitmap *bitmap);

    void TryToStartDrawing(TBool aResumeFromFrontEnd);

    CGameController &GameController() const {return *iGameController;}

    TBool IsKeyPressed(TInt aScanCode);
#if defined __UIQ__ || defined __NOKIA7710__
    void GetJoyStatus(TInt &x, TInt &y, TBool &pressed);
#endif

    void ClearScreen();

    void SetGameScreenDrawer(CGameScreenDrawer *aGameScreenDrawer);

    void Show();
    void Hide();
    void CaptureKeyEvents();
    void CaptureKey(TUint aKeyCode);
    void CaptureKeyUpAndDowns(TUint aScanCode);
    void ReleaseKeyEvents();

    RWsSession& WsSession() {return iWsSession;}

protected:
    CGameWindow();
    void ConstructL(const TDesC& aGameName, TBool aVirtualKeyboardEnabled);
    
    CGameScreenDrawer *iGameScreenDrawer;
    
		/*
		 * Our game device
		 * For target builds it's a CHwFbGameScreenDevice object, for WINS 
		 * a CFbsGameScreenDevice object
		 */
	void ConstructScreenDeviceL();
	void DestructScreenDevice();
	CGameScreenDevice*	iGameScreenDevice;

		/*
		 * Window Server stuff
		 */
	void WsBlitBitmap(CFbsGameScreenDevice* aFbsGameScreenDevice, const TRect& aRect);
	void ConstructWsResourcesL();
	void DestructWsResources();
	CGameDirectScreenAccess*	iGameDirectScreenAccess;
	RWsSession					iWsSession;
	RWindowGroup				iWsWindowGroup;
	RWindow						iWsWindow;
	CWsScreenDevice*			iWsScreen;
	CWindowGc*					iWindowGc;
	TBool						iUseWServ;

	CGameController *iGameController;

	TInt32 iNumberOfCapturedKeys;
	TInt32 iCapturedKeys[KMaxCapturedKeys];
	TInt32 iNumberOfCapturedKeyUpAndDowns;
	TInt32 iCapturedKeyUpAndDowns[KMaxCapturedKeys];

	TBuf<32>	iGameName;

	friend		class CGameDirectScreenAccess;
};


#define MAX_SIMULTANEOUS_KEYPRESSES 5

class CGameController : public CActive
{
 public:
  ~CGameController();
  static CGameController* NewL(RWsSession& aWsSession, 
			       CGameWindow &aGameWindow,
			       TBool aVirtualKeyboardEnabled);

  TBool IsKeyPressed(TInt aScanCode);
#if defined __UIQ__ || defined __NOKIA7710__
  void GetJoyStatus(TInt &x, TInt &y, TBool &pressed);
  void DrawVirtualKeyboard();
#endif
  
 protected:
  CGameController(RWsSession& aWsSession, 
		  CGameWindow &aGameWindow,
		  TBool aVirtualKeyboardEnabled);
  void ConstructL();
  
  void RunL();
  void DoCancel();

 private:
  void HandleKeyEvent(TInt aType, const TKeyEvent& aKeyEvent);
#if defined __UIQ__ || defined __NOKIA7710__
  void HandlePointerEvent(const TPointerEvent &aPointerEvent);
  TBool HandlePossibleVirtualKeyboardPress(const TPointerEvent &aPointerEvent);
#endif
  void KeyDown(TInt aScanCode);
  void KeyUp(TInt aScanCode);
  TBool KeyNeedsAutoUp(TInt aScanCode);

  RWsSession &iWsSession;
  CGameWindow &iGameWindow;
  TInt iScanCodes[MAX_SIMULTANEOUS_KEYPRESSES];
  TInt iAutoUpTimers[MAX_SIMULTANEOUS_KEYPRESSES];
  TInt iNoOfKeyPresses;
  
  TInt iX;
  TInt iY;
  TBool iPressed;

#if defined __UIQ__ || defined __NOKIA7710__
    TBool iVirtualKeyboardEnabled;
    TBool iVirtualKeyboardVisible;
    CFbsBitmap *iVirtualKeyboardBitmap;
    TSize iVirtualKeyboardSize;
    TPoint iVirtualKeyboardPosition;
    TUint8 *iPixelToVirtualButtonMap;
    static const TInt KNoOfVirtualKeys = 16;
    enum TRawVirtualKey
    {
      ERawVirtKeyE,
      ERawVirtKeyNE,
      ERawVirtKeyN,
      ERawVirtKeyNW,
      ERawVirtKeyW,
      ERawVirtKeySW,
      ERawVirtKeyS,
      ERawVirtKeySE,
      ERawVirtKeyCenter,
      ERawVirtKeyFire,
      ERawVirtKeyConf,
      ERawVirtKeySel,
      ERawVirtKeyEsc,
      ERawVirtKeyCoin,
      ERawVirtKeyStart,
      ERawVirtKeyHide,
      ERawVirtKeyNone,
    };
    TRawVirtualKey iCurrentPressedRawVirtualKey;
    TBool iPressOccurredInsideVirtualKeyboard;
    TBool iPressOccurredInsideCurrentVirtualKey;
#endif
};

#if defined __UIQ__ || defined __NOKIA7710__
enum TMameVirtualKeys
{
  EMameVirtKeyUp = 0x300, // Let's hope this doesn't clash with anything!
  EMameVirtKeyDown,
  EMameVirtKeyLeft,
  EMameVirtKeyRight,
  EMameVirtKeyCenter,
  EMameVirtKeyFire,
  EMameVirtKeyConf,
  EMameVirtKeySel,
  EMameVirtKeyEsc,
  EMameVirtKeyCoin,
  EMameVirtKeyStart,
};
#endif

class CGameDirectScreenAccess : public CBase, 
                                public MDirectScreenAccess
{
  public:
    ~CGameDirectScreenAccess();
    static CGameDirectScreenAccess *NewL(CGameWindow& aGameWindow);
    
    void TryToStartDrawing(TBool aResumeFromFrontEnd);
    void AbortNow(RDirectScreenAccess::TTerminationReasons);
    void Restart(RDirectScreenAccess::TTerminationReasons);
    void Update();

protected:
    CGameDirectScreenAccess(CGameWindow& aGameWindow);
    void ConstructL();
    
    CDirectScreenAccess *iDirectScreenAccess;
    RRegion iDirectScreenUpdateRegion;
    CGameWindow &iGameWindow;
};

#endif			/* __VIDEO_H */
